#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""The linear swap network."""

from typing import Callable, cast, Iterable, List, Sequence

import cirq

from openfermion.circuits.gates import FSWAP


def swap_network(
        qubits: Sequence[cirq.Qid],
        operation: Callable[[int, int, cirq.Qid, cirq.Qid], cirq.
                            OP_TREE] = lambda p, q, p_qubit, q_qubit: (),
        fermionic: bool = False,
        offset: bool = False) -> List[cirq.Operation]:
    """Apply operations to pairs of qubits or modes using a swap network.

    This is used for applying operations between arbitrary pairs of qubits or
    fermionic modes using only nearest-neighbor interactions on a linear array
    of qubits. It works by reversing the order of qubits or modes with a
    sequence of swap gates and applying an operation when the relevant qubits
    or modes become adjacent. For fermionic modes, this assumes the
    Jordan-Wigner Transform.

    Examples
    --------

    Input:

    .. testcode::

        import cirq
        from openfermion import swap_network

        qubits = cirq.LineQubit.range(4)
        circuit = cirq.Circuit(swap_network(qubits))
        print(circuit)

    Output:

    .. testoutput::

        0: ───×───────×───────
              │       │
        1: ───×───×───×───×───
                  │       │
        2: ───×───×───×───×───
              │       │
        3: ───×───────×───────

    Input:

    .. testcode::

        circuit = cirq.Circuit(swap_network(qubits, offset=True))
        print(circuit)

    Output:

    .. testoutput::

        0: ───────×───────×───
                  │       │
        1: ───×───×───×───×───
              │       │
        2: ───×───×───×───×───
                  │       │
        3: ───────×───────×───

    Input:

    .. testcode::

        circuit = cirq.Circuit(
            swap_network(
                qubits,
                lambda p, q, a, b: cirq.ISWAP(a, b)**-1 if abs(p - q) == 1
                                   else cirq.CZ(a, b),
                fermionic=True))
        print(circuit)

    Output:

    .. testoutput::

        0: ───iSwap──────×ᶠ────────────@───×ᶠ───────────────────
              │          │             │   │
        1: ───iSwap^-1───×ᶠ───@───×ᶠ───@───×ᶠ───iSwap──────×ᶠ───
                              │   │             │          │
        2: ───iSwap──────×ᶠ───@───×ᶠ───@───×ᶠ───iSwap^-1───×ᶠ───
              │          │             │   │
        3: ───iSwap^-1───×ᶠ────────────@───×ᶠ───────────────────


    Args:
        qubits: The qubits sorted so that the j-th qubit in the Sequence
            represents the j-th qubit or fermionic mode.
        operation: Returns extra interactions to perform between qubits/modes as
            they are swapped past each other. A call to this function takes the
            form ``operation(p, q, p_qubit, q_qubit)`` where p and q are indices
            representing either qubits or fermionic modes, and p_qubit and
            q_qubit are the qubits which are currently storing those modes.
        fermionic: If True, use fermionic swaps under the JWT (that is, swap
            fermionic modes instead of qubits). If False, use normal qubit
            swaps.
        offset: If True, then qubit 0 will participate in odd-numbered layers
            instead of even-numbered layers.
    """
    n_qubits = len(qubits)
    order = list(range(n_qubits))
    swap_gate = FSWAP if fermionic else cirq.SWAP
    result = []  # type: List[cirq.Operation]

    for layer_num in range(n_qubits):
        lowest_active_qubit = (layer_num + offset) % 2
        active_pairs = (
            (i, i + 1) for i in range(lowest_active_qubit, n_qubits - 1, 2))
        for i, j in active_pairs:
            p, q = order[i], order[j]
            extra_ops = operation(p, q, qubits[i], qubits[j])
            result.extend(
                cast(Iterable[cirq.Operation], cirq.flatten_op_tree(extra_ops)))
            result.append(swap_gate(qubits[i], qubits[j]))
            order[i], order[j] = q, p

    return result



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Useful miscelaneous functions to transform QubitOperators
"""

import numpy

from openfermion.ops.operators import QubitOperator


def project_onto_sector(operator, qubits, sectors):
    """
    Remove qubit by projecting onto sector.

    Takes a QubitOperator, and projects out a list
    of qubits, into either the +1 or -1 sector.
    Note - this requires knowledge of which sector
    we wish to project into.

    Args:
        operator: the QubitOperator to work on
        qubits: a list of indices of qubits in
           operator to remove
        sectors: for each qubit, whether to project
            into the 0 subspace (<Z>=1) or the
            1 subspace (<Z>=-1).

    Returns:
        projected_operator: the resultant operator

    Raises:
        TypeError: operator must be a QubitOperator.
        TypeError: qubits and sector must be an array-like.
        ValueError: If qubits and sectors have different length.
        ValueError: If sector are not specified as 0 or 1.
    """
    if not isinstance(operator, QubitOperator):
        raise TypeError('Input operator must be a QubitOperator.')
    if not isinstance(qubits, (list, numpy.ndarray)):
        raise TypeError('Qubit input must be an array-like.')
    if not isinstance(sectors, (list, numpy.ndarray)):
        raise TypeError('Sector input must be an array-like.')
    if len(qubits) != len(sectors):
        raise ValueError('Number of qubits and sectors must be equal.')
    for i in sectors:
        if i not in [0, 1]:
            raise ValueError('Sectors must be 0 or 1.')

    projected_operator = QubitOperator()
    for term, factor in operator.terms.items():

        # Any term containing X or Y on the removed
        # qubits has an expectation value of zero
        if [t for t in term if t[0] in qubits and t[1] in ['X', 'Y']]:
            continue

        new_term = tuple((t[0] - len([q for q in qubits if q < t[0]]),
                          t[1]) for t in term if t[0] not in qubits)
        new_factor =\
            factor * (-1)**(sum([sectors[qubits.index(t[0])]
                                 for t in term if t[0] in qubits]))
        projected_operator += QubitOperator(new_term, new_factor)

    return projected_operator


def projection_error(operator, qubits, sectors):
    """
    Calculate the error from the project_onto_sector function.

    Args:
        operator: the QubitOperator to work on
        qubits: a list of indices of qubits in
           operator to remove
        sectors: for each qubit, whether to project
            into the 0 subspace (<Z>=1) or the
            1 subspace (<Z>=-1).

    Returns:
        error: the trace norm of the removed term.

    Raises:
        TypeError: operator must be a QubitOperator.
        TypeError: qubits and sector must be an array-like.
        ValueError: If qubits and sectors have different length.
        ValueError: If sector are not specified as 0 or 1.
    """
    if not isinstance(operator, QubitOperator):
        raise TypeError('Input operator must be a QubitOperator.')
    if not isinstance(qubits, (list, numpy.ndarray)):
        raise TypeError('Qubit input must be an array-like.')
    if not isinstance(sectors, (list, numpy.ndarray)):
        raise TypeError('Sector input must be an array-like.')
    if len(qubits) != len(sectors):
        raise ValueError('Number of qubits and sectors must be equal.')
    for i in sectors:
        if i not in [0, 1]:
            raise ValueError('Sectors must be 0 or 1.')

    error = 0
    for term, factor in operator.terms.items():

        # Any term containing X or Y on the removed
        # qubits contributes to the error
        if [t for t in term if t[0] in qubits and t[1] in ['X', 'Y']]:
            error += abs(factor)**2

    return numpy.sqrt(error)


def rotate_qubit_by_pauli(qop, pauli, angle):
    r"""
    Rotate qubit operator by exponential of Pauli.

    Perform the rotation e^{-i \theta * P}Qe^{i \theta * P}
    on a qubitoperator Q and a Pauli operator P.

    Args:
        qop: the QubitOperator to be rotated
        pauli: a single Pauli operator - a QubitOperator with
            a single term, and a coefficient equal to 1.
        angle: the angle to be rotated by.

    Returns:
        rotated_op - the rotated QubitOperator following the
            above formula.

    Raises:
        TypeError: qop must be a QubitOperator
        TypeError: pauli must be a Pauli Operator (QubitOperator
            with single term and coefficient equal to 1).
    """
    pvals = list(pauli.terms.values())
    if type(qop) is not QubitOperator:
        raise TypeError('This can only rotate QubitOperators')

    if (type(pauli) is not QubitOperator or len(pauli.terms) != 1 or
            pvals[0] != 1):
        raise TypeError('This can only rotate by Pauli operators')

    pqp = pauli * qop * pauli
    even_terms = 0.5 * (qop + pqp)
    odd_terms = 0.5 * (qop - pqp)

    rotated_op = even_terms + numpy.cos(2 * angle) * odd_terms + \
        1j * numpy.sin(2 * angle) * odd_terms * pauli

    return rotated_op



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from .general_hubbard import (
    FermiHubbardModel,
    number_operator,
    interaction_operator,
    tunneling_operator,
    number_difference_operator,
)

from .hartree_fock import (
    rhf_func_generator,
    rhf_minimization,
    HartreeFockFunctional,
    rhf_params_to_matrix,
    get_matrix_of_eigs,
    generate_hamiltonian,
)

from .hubbard import (
    bose_hubbard,
    fermi_hubbard,
)

from .jellium import (
    dual_basis_kinetic,
    dual_basis_potential,
    dual_basis_jellium_model,
    jellium_model,
    jordan_wigner_dual_basis_jellium,
    hypercube_grid_with_given_wigner_seitz_radius_and_filling,
    plane_wave_kinetic,
    plane_wave_potential,
    wigner_seitz_length_scale,
)

from .jellium_hf_state import (
    hartree_fock_state_jellium,
    lowest_single_particle_energy_states,
)

from .mean_field_dwave import mean_field_dwave

from .plane_wave_hamiltonian import (
    dual_basis_external_potential,
    plane_wave_external_potential,
    plane_wave_hamiltonian,
    jordan_wigner_dual_basis_hamiltonian,
)

from .special_operators import (
    s_plus_operator,
    s_squared_operator,
    sx_operator,
    sy_operator,
    sz_operator,
    majorana_operator,
    number_operator,
)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for sparse_tools.py."""
import os

import unittest

import itertools

import numpy
from numpy.linalg import multi_dot

from scipy.linalg import eigh, norm
from scipy.sparse import csc_matrix
import scipy
from scipy.special import comb

from openfermion.config import DATA_DIRECTORY
from openfermion.chem import MolecularData
from openfermion.hamiltonians import (fermi_hubbard, jellium_model,
                                      wigner_seitz_length_scale,
                                      number_operator)
from openfermion.ops.operators import (FermionOperator, BosonOperator,
                                       QuadOperator, QubitOperator,
                                       MajoranaOperator)
from openfermion.ops.representations import DiagonalCoulombHamiltonian
from openfermion.transforms.opconversions import (jordan_wigner,
                                                  get_fermion_operator,
                                                  normal_ordered)
from openfermion.transforms.repconversions import (fourier_transform,
                                                   get_interaction_operator)

from openfermion.testing.testing_utils import random_hermitian_matrix
from openfermion.linalg.sparse_tools import (
    get_sparse_operator, get_ground_state, eigenspectrum, expectation,
    jw_number_restrict_state, inner_product, jw_sz_restrict_state,
    jw_get_ground_state_at_particle_number, jw_sparse_givens_rotation,
    pauli_matrix_map, sparse_eigenspectrum, jordan_wigner_sparse,
    kronecker_operators, identity_csc, pauli_x_csc, qubit_operator_sparse,
    get_linear_qubit_operator_diagonal, jw_number_indices,
    get_number_preserving_sparse_operator, jw_configuration_state,
    jw_hartree_fock_state, jw_sz_restrict_operator, jw_sz_indices,
    jw_number_restrict_operator, variance,
    expectation_computational_basis_state,
    expectation_db_operator_with_pw_basis_state, get_gap, boson_ladder_sparse,
    boson_operator_sparse, single_quad_op_sparse, _iterate_basis_)

from openfermion.utils.operator_utils import (is_hermitian, count_qubits,
                                              hermitian_conjugated)
from openfermion.utils.indexing import up_index, down_index
from openfermion.utils.grid import Grid
from openfermion.hamiltonians.jellium_hf_state import (
    lowest_single_particle_energy_states)
from openfermion.linalg.linear_qubit_operator import LinearQubitOperator


class EigenSpectrumTest(unittest.TestCase):

    def setUp(self):
        self.n_qubits = 5
        self.majorana_operator = MajoranaOperator((1, 4, 9))
        self.fermion_term = FermionOperator('1^ 2^ 3 4', -3.17)
        self.fermion_operator = self.fermion_term + hermitian_conjugated(
            self.fermion_term)
        self.qubit_operator = jordan_wigner(self.fermion_operator)
        self.interaction_operator = get_interaction_operator(
            self.fermion_operator)

    def test_eigenspectrum(self):
        fermion_eigenspectrum = eigenspectrum(self.fermion_operator)
        qubit_eigenspectrum = eigenspectrum(self.qubit_operator)
        interaction_eigenspectrum = eigenspectrum(self.interaction_operator)
        for i in range(2**self.n_qubits):
            self.assertAlmostEqual(fermion_eigenspectrum[i],
                                   qubit_eigenspectrum[i])
            self.assertAlmostEqual(fermion_eigenspectrum[i],
                                   interaction_eigenspectrum[i])

        with self.assertRaises(TypeError):
            _ = eigenspectrum(BosonOperator())

        with self.assertRaises(TypeError):
            _ = eigenspectrum(QuadOperator())


class SparseOperatorTest(unittest.TestCase):

    def test_kronecker_operators(self):

        self.assertAlmostEqual(
            0,
            numpy.amax(
                numpy.absolute(
                    kronecker_operators(3 * [identity_csc]) -
                    kronecker_operators(3 * [pauli_x_csc])**2)))

    def test_qubit_jw_fermion_integration(self):

        # Initialize a random fermionic operator.
        fermion_operator = FermionOperator(((3, 1), (2, 1), (1, 0), (0, 0)),
                                           -4.3)
        fermion_operator += FermionOperator(((3, 1), (1, 0)), 8.17)
        fermion_operator += 3.2 * FermionOperator()

        # Map to qubits and compare matrix versions.
        qubit_operator = jordan_wigner(fermion_operator)
        qubit_sparse = get_sparse_operator(qubit_operator)
        qubit_spectrum = sparse_eigenspectrum(qubit_sparse)
        fermion_sparse = jordan_wigner_sparse(fermion_operator)
        fermion_spectrum = sparse_eigenspectrum(fermion_sparse)
        self.assertAlmostEqual(
            0., numpy.amax(numpy.absolute(fermion_spectrum - qubit_spectrum)))


class JordanWignerSparseTest(unittest.TestCase):

    def test_jw_sparse_0create(self):
        expected = csc_matrix(([1], ([1], [0])), shape=(2, 2))
        self.assertTrue(
            numpy.allclose(
                jordan_wigner_sparse(FermionOperator('0^')).A, expected.A))

    def test_jw_sparse_1annihilate(self):
        expected = csc_matrix(([1, -1], ([0, 2], [1, 3])), shape=(4, 4))
        self.assertTrue(
            numpy.allclose(
                jordan_wigner_sparse(FermionOperator('1')).A, expected.A))

    def test_jw_sparse_0create_2annihilate(self):
        expected = csc_matrix(([-1j, 1j], ([4, 6], [1, 3])),
                              shape=(8, 8),
                              dtype=numpy.complex128)
        self.assertTrue(
            numpy.allclose(
                jordan_wigner_sparse(FermionOperator('0^ 2', -1j)).A,
                expected.A))

    def test_jw_sparse_0create_3annihilate(self):
        expected = csc_matrix(
            ([-1j, 1j, 1j, -1j], ([8, 10, 12, 14], [1, 3, 5, 7])),
            shape=(16, 16),
            dtype=numpy.complex128)
        self.assertTrue(
            numpy.allclose(
                jordan_wigner_sparse(FermionOperator('0^ 3', -1j)).A,
                expected.A))

    def test_jw_sparse_twobody(self):
        expected = csc_matrix(([1, 1], ([6, 14], [5, 13])), shape=(16, 16))
        self.assertTrue(
            numpy.allclose(
                jordan_wigner_sparse(FermionOperator('2^ 1^ 1 3')).A,
                expected.A))

    def test_qubit_operator_sparse_n_qubits_too_small(self):
        with self.assertRaises(ValueError):
            qubit_operator_sparse(QubitOperator('X3'), 1)

    def test_qubit_operator_sparse_n_qubits_not_specified(self):
        expected = csc_matrix(([1, 1, 1, 1], ([1, 0, 3, 2], [0, 1, 2, 3])),
                              shape=(4, 4))
        self.assertTrue(
            numpy.allclose(
                qubit_operator_sparse(QubitOperator('X1')).A, expected.A))

    def test_get_linear_qubit_operator_diagonal_wrong_n(self):
        """Testing with wrong n_qubits."""
        with self.assertRaises(ValueError):
            get_linear_qubit_operator_diagonal(QubitOperator('X3'), 1)

    def test_get_linear_qubit_operator_diagonal_0(self):
        """Testing with zero term."""
        qubit_operator = QubitOperator.zero()
        vec_expected = numpy.zeros(8)

        self.assertTrue(
            numpy.allclose(
                get_linear_qubit_operator_diagonal(qubit_operator, 3),
                vec_expected))

    def test_get_linear_qubit_operator_diagonal_zero(self):
        """Get zero diagonals from get_linear_qubit_operator_diagonal."""
        qubit_operator = QubitOperator('X0 Y1')
        vec_expected = numpy.zeros(4)

        self.assertTrue(
            numpy.allclose(get_linear_qubit_operator_diagonal(qubit_operator),
                           vec_expected))

    def test_get_linear_qubit_operator_diagonal_non_zero(self):
        """Get non zero diagonals from get_linear_qubit_operator_diagonal."""
        qubit_operator = QubitOperator('Z0 Z2')
        vec_expected = numpy.array([1, -1, 1, -1, -1, 1, -1, 1])

        self.assertTrue(
            numpy.allclose(get_linear_qubit_operator_diagonal(qubit_operator),
                           vec_expected))

    def test_get_linear_qubit_operator_diagonal_cmp_zero(self):
        """Compare get_linear_qubit_operator_diagonal with
            get_linear_qubit_operator."""
        qubit_operator = QubitOperator('Z1 X2 Y5')
        vec_expected = numpy.diag(
            LinearQubitOperator(qubit_operator) * numpy.eye(2**6))

        self.assertTrue(
            numpy.allclose(get_linear_qubit_operator_diagonal(qubit_operator),
                           vec_expected))

    def test_get_linear_qubit_operator_diagonal_cmp_non_zero(self):
        """Compare get_linear_qubit_operator_diagonal with
            get_linear_qubit_operator."""
        qubit_operator = QubitOperator('Z1 Z2 Z5')
        vec_expected = numpy.diag(
            LinearQubitOperator(qubit_operator) * numpy.eye(2**6))

        self.assertTrue(
            numpy.allclose(get_linear_qubit_operator_diagonal(qubit_operator),
                           vec_expected))


class ComputationalBasisStateTest(unittest.TestCase):

    def test_computational_basis_state(self):
        comp_basis_state = jw_configuration_state([0, 2, 5], 7)
        self.assertAlmostEqual(comp_basis_state[82], 1.)
        self.assertAlmostEqual(sum(comp_basis_state), 1.)


class JWHartreeFockStateTest(unittest.TestCase):

    def test_jw_hartree_fock_state(self):
        hartree_fock_state = jw_hartree_fock_state(3, 7)
        self.assertAlmostEqual(hartree_fock_state[112], 1.)
        self.assertAlmostEqual(sum(hartree_fock_state), 1.)


class JWNumberIndicesTest(unittest.TestCase):

    def test_jw_sparse_index(self):
        """Test the indexing scheme for selecting specific particle numbers"""
        expected = [1, 2]
        calculated_indices = jw_number_indices(1, 2)
        self.assertEqual(expected, calculated_indices)

        expected = [3]
        calculated_indices = jw_number_indices(2, 2)
        self.assertEqual(expected, calculated_indices)

    def test_jw_number_indices(self):
        n_qubits = numpy.random.randint(1, 12)
        n_particles = numpy.random.randint(n_qubits + 1)

        number_indices = jw_number_indices(n_particles, n_qubits)
        subspace_dimension = len(number_indices)

        self.assertEqual(subspace_dimension, comb(n_qubits, n_particles))

        for index in number_indices:
            binary_string = bin(index)[2:].zfill(n_qubits)
            n_ones = binary_string.count('1')
            self.assertEqual(n_ones, n_particles)


class JWSzIndicesTest(unittest.TestCase):

    def test_jw_sz_indices(self):
        """Test the indexing scheme for selecting specific sz value"""

        def sz_integer(bitstring):
            """Computes the total number of occupied up sites
            minus the total number of occupied down sites."""
            n_sites = len(bitstring) // 2

            n_up = len([
                site for site in range(n_sites)
                if bitstring[up_index(site)] == '1'
            ])
            n_down = len([
                site for site in range(n_sites)
                if bitstring[down_index(site)] == '1'
            ])

            return n_up - n_down

        def jw_sz_indices_brute_force(sz_value, n_qubits):
            """Computes the correct indices by brute force."""
            indices = []
            for bitstring in itertools.product(['0', '1'], repeat=n_qubits):
                if (sz_integer(bitstring) == int(2 * sz_value)):
                    indices.append(int(''.join(bitstring), 2))

            return indices

        # General test
        n_sites = numpy.random.randint(1, 10)
        n_qubits = 2 * n_sites
        sz_int = ((-1)**numpy.random.randint(2) *
                  numpy.random.randint(n_sites + 1))
        sz_value = sz_int / 2.

        correct_indices = jw_sz_indices_brute_force(sz_value, n_qubits)
        subspace_dimension = len(correct_indices)

        calculated_indices = jw_sz_indices(sz_value, n_qubits)

        self.assertEqual(len(calculated_indices), subspace_dimension)

        for index in calculated_indices:
            binary_string = bin(index)[2:].zfill(n_qubits)
            self.assertEqual(sz_integer(binary_string), sz_int)

        # Test fixing particle number
        n_particles = abs(sz_int)

        correct_indices = [
            index for index in correct_indices
            if bin(index)[2:].count('1') == n_particles
        ]
        subspace_dimension = len(correct_indices)

        calculated_indices = jw_sz_indices(sz_value,
                                           n_qubits,
                                           n_electrons=n_particles)

        self.assertEqual(len(calculated_indices), subspace_dimension)

        for index in calculated_indices:
            binary_string = bin(index)[2:].zfill(n_qubits)
            self.assertEqual(sz_integer(binary_string), sz_int)
            self.assertEqual(binary_string.count('1'), n_particles)

        # Test exceptions
        with self.assertRaises(ValueError):
            jw_sz_indices(3, 3)

        with self.assertRaises(ValueError):
            jw_sz_indices(3.1, 4)

        with self.assertRaises(ValueError):
            jw_sz_indices(1.5, 8, n_electrons=6)

        with self.assertRaises(ValueError):
            jw_sz_indices(1.5, 8, n_electrons=1)


class JWNumberRestrictOperatorTest(unittest.TestCase):

    def test_jw_restrict_operator(self):
        """Test the scheme for restricting JW encoded operators to number"""
        # Make a Hamiltonian that cares mostly about number of electrons
        n_qubits = 4
        target_electrons = 2
        penalty_const = 10.
        number_sparse = jordan_wigner_sparse(number_operator(n_qubits))
        bias_sparse = jordan_wigner_sparse(
            sum([
                FermionOperator(((i, 1), (i, 0)), 1.0) for i in range(n_qubits)
            ], FermionOperator()))
        hamiltonian_sparse = penalty_const * (
            number_sparse -
            target_electrons * scipy.sparse.identity(2**n_qubits)
        ).dot(number_sparse - target_electrons *
              scipy.sparse.identity(2**n_qubits)) + bias_sparse

        restricted_hamiltonian = jw_number_restrict_operator(
            hamiltonian_sparse, target_electrons, n_qubits)
        true_eigvals, _ = eigh(hamiltonian_sparse.A)
        test_eigvals, _ = eigh(restricted_hamiltonian.A)

        self.assertAlmostEqual(norm(true_eigvals[:6] - test_eigvals[:6]), 0.0)

    def test_jw_restrict_operator_hopping_to_1_particle(self):
        hop = FermionOperator('3^ 1') + FermionOperator('1^ 3')
        hop_sparse = jordan_wigner_sparse(hop, n_qubits=4)
        hop_restrict = jw_number_restrict_operator(hop_sparse, 1, n_qubits=4)
        expected = csc_matrix(([1, 1], ([0, 2], [2, 0])), shape=(4, 4))

        self.assertTrue(numpy.allclose(hop_restrict.A, expected.A))

    def test_jw_restrict_operator_interaction_to_1_particle(self):
        interaction = FermionOperator('3^ 2^ 4 1')
        interaction_sparse = jordan_wigner_sparse(interaction, n_qubits=6)
        interaction_restrict = jw_number_restrict_operator(interaction_sparse,
                                                           1,
                                                           n_qubits=6)
        expected = csc_matrix(([], ([], [])), shape=(6, 6))

        self.assertTrue(numpy.allclose(interaction_restrict.A, expected.A))

    def test_jw_restrict_operator_interaction_to_2_particles(self):
        interaction = (FermionOperator('3^ 2^ 4 1') +
                       FermionOperator('4^ 1^ 3 2'))
        interaction_sparse = jordan_wigner_sparse(interaction, n_qubits=6)
        interaction_restrict = jw_number_restrict_operator(interaction_sparse,
                                                           2,
                                                           n_qubits=6)

        dim = 6 * 5 // 2  # shape of new sparse array

        # 3^ 2^ 4 1 maps 2**4 + 2 = 18 to 2**3 + 2**2 = 12 and vice versa;
        # in the 2-particle subspace (1, 4) and (2, 3) are 7th and 9th.
        expected = csc_matrix(([-1, -1], ([7, 9], [9, 7])), shape=(dim, dim))

        self.assertTrue(numpy.allclose(interaction_restrict.A, expected.A))

    def test_jw_restrict_operator_hopping_to_1_particle_default_nqubits(self):
        interaction = (FermionOperator('3^ 2^ 4 1') +
                       FermionOperator('4^ 1^ 3 2'))
        interaction_sparse = jordan_wigner_sparse(interaction, n_qubits=6)
        # n_qubits should default to 6
        interaction_restrict = jw_number_restrict_operator(
            interaction_sparse, 2)

        dim = 6 * 5 // 2  # shape of new sparse array

        # 3^ 2^ 4 1 maps 2**4 + 2 = 18 to 2**3 + 2**2 = 12 and vice versa;
        # in the 2-particle subspace (1, 4) and (2, 3) are 7th and 9th.
        expected = csc_matrix(([-1, -1], ([7, 9], [9, 7])), shape=(dim, dim))

        self.assertTrue(numpy.allclose(interaction_restrict.A, expected.A))

    def test_jw_restrict_jellium_ground_state_integration(self):
        n_qubits = 4
        grid = Grid(dimensions=1, length=n_qubits, scale=1.0)
        jellium_hamiltonian = jordan_wigner_sparse(
            jellium_model(grid, spinless=False))

        #  2 * n_qubits because of spin
        number_sparse = jordan_wigner_sparse(number_operator(2 * n_qubits))

        restricted_number = jw_number_restrict_operator(number_sparse, 2)
        restricted_jellium_hamiltonian = jw_number_restrict_operator(
            jellium_hamiltonian, 2)

        _, ground_state = get_ground_state(restricted_jellium_hamiltonian)

        number_expectation = expectation(restricted_number, ground_state)
        self.assertAlmostEqual(number_expectation, 2)


class JWSzRestrictOperatorTest(unittest.TestCase):

    def test_restrict_interaction_hamiltonian(self):
        """Test restricting a coulomb repulsion Hamiltonian to a specified
        Sz manifold."""
        x_dim = 3
        y_dim = 2

        interaction_term = fermi_hubbard(x_dim, y_dim, 0., 1.)
        interaction_sparse = get_sparse_operator(interaction_term)
        sz_value = 2
        interaction_restricted = jw_sz_restrict_operator(
            interaction_sparse, sz_value)
        restricted_interaction_values = set(
            [int(value.real) for value in interaction_restricted.diagonal()])
        # Originally the eigenvalues run from 0 to 6 but after restricting,
        # they should run from 0 to 2
        self.assertEqual(restricted_interaction_values, {0, 1, 2})


class JWNumberRestrictStateTest(unittest.TestCase):

    def test_jw_number_restrict_state(self):
        n_qubits = numpy.random.randint(1, 12)
        n_particles = numpy.random.randint(0, n_qubits)

        number_indices = jw_number_indices(n_particles, n_qubits)
        subspace_dimension = len(number_indices)

        # Create a vector that has entry 1 for every coordinate with
        # the specified particle number, and 0 everywhere else
        vector = numpy.zeros(2**n_qubits, dtype=float)
        vector[number_indices] = 1

        # Restrict the vector
        restricted_vector = jw_number_restrict_state(vector, n_particles)

        # Check that it has the correct shape
        self.assertEqual(restricted_vector.shape[0], subspace_dimension)

        # Check that it has the same norm as the original vector
        self.assertAlmostEqual(
            inner_product(vector, vector),
            inner_product(restricted_vector, restricted_vector))


class JWSzRestrictStateTest(unittest.TestCase):

    def test_jw_sz_restrict_state(self):
        n_sites = numpy.random.randint(1, 10)
        n_qubits = 2 * n_sites
        sz_int = ((-1)**numpy.random.randint(2) *
                  numpy.random.randint(n_sites + 1))
        sz_value = sz_int / 2

        sz_indices = jw_sz_indices(sz_value, n_qubits)
        subspace_dimension = len(sz_indices)

        # Create a vector that has entry 1 for every coordinate in
        # the specified subspace, and 0 everywhere else
        vector = numpy.zeros(2**n_qubits, dtype=float)
        vector[sz_indices] = 1

        # Restrict the vector
        restricted_vector = jw_sz_restrict_state(vector, sz_value)

        # Check that it has the correct shape
        self.assertEqual(restricted_vector.shape[0], subspace_dimension)

        # Check that it has the same norm as the original vector
        self.assertAlmostEqual(
            inner_product(vector, vector),
            inner_product(restricted_vector, restricted_vector))


class JWGetGroundStatesByParticleNumberTest(unittest.TestCase):

    def test_jw_get_ground_state_at_particle_number_herm_conserving(self):
        # Initialize a particle-number-conserving Hermitian operator
        ferm_op = FermionOperator('0^ 1') + FermionOperator('1^ 0') + \
            FermionOperator('1^ 2') + FermionOperator('2^ 1') + \
            FermionOperator('1^ 3', -.4) + FermionOperator('3^ 1', -.4)
        jw_hamiltonian = jordan_wigner(ferm_op)
        sparse_operator = get_sparse_operator(jw_hamiltonian)
        n_qubits = 4

        num_op = get_sparse_operator(number_operator(n_qubits))

        # Test each possible particle number
        for particle_number in range(n_qubits):
            # Get the ground energy and ground state at this particle number
            energy, state = jw_get_ground_state_at_particle_number(
                sparse_operator, particle_number)

            # Check that it's an eigenvector with the correct eigenvalue
            self.assertTrue(
                numpy.allclose(sparse_operator.dot(state), energy * state))

            # Check that it has the correct particle number
            num = expectation(num_op, state)
            self.assertAlmostEqual(num, particle_number)

    def test_jw_get_ground_state_at_particle_number_hubbard(self):

        model = fermi_hubbard(2, 2, 1.0, 4.0)
        sparse_operator = get_sparse_operator(model)
        n_qubits = count_qubits(model)
        num_op = get_sparse_operator(number_operator(n_qubits))

        # Test each possible particle number
        for particle_number in range(n_qubits):
            # Get the ground energy and ground state at this particle number
            energy, state = jw_get_ground_state_at_particle_number(
                sparse_operator, particle_number)

            # Check that it's an eigenvector with the correct eigenvalue
            self.assertTrue(
                numpy.allclose(sparse_operator.dot(state), energy * state))

            # Check that it has the correct particle number
            num = expectation(num_op, state)
            self.assertAlmostEqual(num, particle_number)

    def test_jw_get_ground_state_at_particle_number_jellium(self):

        grid = Grid(2, 2, 1.0)
        model = jellium_model(grid, spinless=True, plane_wave=False)
        sparse_operator = get_sparse_operator(model)
        n_qubits = count_qubits(model)
        num_op = get_sparse_operator(number_operator(n_qubits))

        # Test each possible particle number
        for particle_number in range(n_qubits):
            # Get the ground energy and ground state at this particle number
            energy, state = jw_get_ground_state_at_particle_number(
                sparse_operator, particle_number)

            # Check that it's an eigenvector with the correct eigenvalue
            self.assertTrue(
                numpy.allclose(sparse_operator.dot(state), energy * state))

            # Check that it has the correct particle number
            num = expectation(num_op, state)
            self.assertAlmostEqual(num, particle_number)


class JWSparseGivensRotationTest(unittest.TestCase):

    def test_bad_input(self):
        with self.assertRaises(ValueError):
            jw_sparse_givens_rotation(0, 2, 1., 1., 5)
        with self.assertRaises(ValueError):
            jw_sparse_givens_rotation(4, 5, 1., 1., 5)


class GroundStateTest(unittest.TestCase):

    def test_get_ground_state_hermitian(self):
        ground = get_ground_state(
            get_sparse_operator(
                QubitOperator('Y0 X1') + QubitOperator('Z0 Z1')))
        expected_state = csc_matrix(([1j, 1], ([1, 2], [0, 0])),
                                    shape=(4, 1),
                                    dtype=numpy.complex128).A
        expected_state /= numpy.sqrt(2.0)

        self.assertAlmostEqual(ground[0], -2)
        self.assertAlmostEqual(
            numpy.absolute(expected_state.T.conj().dot(ground[1]))[0], 1.)


class ExpectationTest(unittest.TestCase):

    def test_expectation_correct_sparse_matrix(self):
        operator = get_sparse_operator(QubitOperator('X0'), n_qubits=2)
        vector = numpy.array([0., 1.j, 0., 1.j])
        self.assertAlmostEqual(expectation(operator, vector), 2.0)

        density_matrix = scipy.sparse.csc_matrix(
            numpy.outer(vector, numpy.conjugate(vector)))
        self.assertAlmostEqual(expectation(operator, density_matrix), 2.0)

    def test_expectation_correct_linear_operator(self):
        operator = LinearQubitOperator(QubitOperator('X0'), n_qubits=2)
        vector = numpy.array([0., 1.j, 0., 1.j])
        self.assertAlmostEqual(expectation(operator, vector), 2.0)

    def test_expectation_handles_column_vector(self):
        operator = get_sparse_operator(QubitOperator('X0'), n_qubits=2)
        vector = numpy.array([[0.], [1.j], [0.], [1.j]])
        self.assertAlmostEqual(expectation(operator, vector), 2.0)

    def test_expectation_correct_zero(self):
        operator = get_sparse_operator(QubitOperator('X0'), n_qubits=2)
        vector = numpy.array([1j, -1j, -1j, -1j])
        self.assertAlmostEqual(expectation(operator, vector), 0.0)

    def test_execptions(self):
        operator = LinearQubitOperator(QubitOperator('X0'), n_qubits=2)
        vector = scipy.sparse.csc_matrix(numpy.array([1, 0, 0, 0]))
        with self.assertRaises(ValueError):
            _ = expectation(operator, vector)
        with self.assertRaises(ValueError):
            _ = expectation(vector, operator)


class VarianceTest(unittest.TestCase):

    def test_variance_row_vector(self):

        X = pauli_matrix_map['X']
        Z = pauli_matrix_map['Z']
        zero = numpy.array([1., 0.])
        plus = numpy.array([1., 1.]) / numpy.sqrt(2)
        minus = numpy.array([1., -1.]) / numpy.sqrt(2)

        self.assertAlmostEqual(variance(Z, zero), 0.)
        self.assertAlmostEqual(variance(X, zero), 1.)

        self.assertAlmostEqual(variance(Z, plus), 1.)
        self.assertAlmostEqual(variance(X, plus), 0.)

        self.assertAlmostEqual(variance(Z, minus), 1.)
        self.assertAlmostEqual(variance(X, minus), 0.)

    def test_variance_column_vector(self):

        X = pauli_matrix_map['X']
        Z = pauli_matrix_map['Z']
        zero = numpy.array([[1.], [0.]])
        plus = numpy.array([[1.], [1.]]) / numpy.sqrt(2)
        minus = numpy.array([[1.], [-1.]]) / numpy.sqrt(2)

        self.assertAlmostEqual(variance(Z, zero), 0.)
        self.assertAlmostEqual(variance(X, zero), 1.)

        self.assertAlmostEqual(variance(Z, plus), 1.)
        self.assertAlmostEqual(variance(X, plus), 0.)

        self.assertAlmostEqual(variance(Z, minus), 1.)
        self.assertAlmostEqual(variance(X, minus), 0.)


class ExpectationComputationalBasisStateTest(unittest.TestCase):

    def test_expectation_fermion_operator_single_number_terms(self):
        operator = FermionOperator('3^ 3', 1.9) + FermionOperator('2^ 1')
        state = csc_matrix(([1], ([15], [0])), shape=(16, 1))

        self.assertAlmostEqual(
            expectation_computational_basis_state(operator, state), 1.9)

    def test_expectation_fermion_operator_two_number_terms(self):
        operator = (FermionOperator('2^ 2', 1.9) + FermionOperator('2^ 1') +
                    FermionOperator('2^ 1^ 2 1', -1.7))
        state = csc_matrix(([1], ([6], [0])), shape=(16, 1))

        self.assertAlmostEqual(
            expectation_computational_basis_state(operator, state), 3.6)

    def test_expectation_identity_fermion_operator(self):
        operator = FermionOperator.identity() * 1.1
        state = csc_matrix(([1], ([6], [0])), shape=(16, 1))

        self.assertAlmostEqual(
            expectation_computational_basis_state(operator, state), 1.1)

    def test_expectation_state_is_list_single_number_terms(self):
        operator = FermionOperator('3^ 3', 1.9) + FermionOperator('2^ 1')
        state = [1, 1, 1, 1]

        self.assertAlmostEqual(
            expectation_computational_basis_state(operator, state), 1.9)

    def test_expectation_state_is_list_fermion_operator_two_number_terms(self):
        operator = (FermionOperator('2^ 2', 1.9) + FermionOperator('2^ 1') +
                    FermionOperator('2^ 1^ 2 1', -1.7))
        state = [0, 1, 1]

        self.assertAlmostEqual(
            expectation_computational_basis_state(operator, state), 3.6)

    def test_expectation_state_is_list_identity_fermion_operator(self):
        operator = FermionOperator.identity() * 1.1
        state = [0, 1, 1]

        self.assertAlmostEqual(
            expectation_computational_basis_state(operator, state), 1.1)

    def test_expectation_bad_operator_type(self):
        with self.assertRaises(TypeError):
            expectation_computational_basis_state(
                'never', csc_matrix(([1], ([6], [0])), shape=(16, 1)))

    def test_expectation_qubit_operator_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            expectation_computational_basis_state(
                QubitOperator(), csc_matrix(([1], ([6], [0])), shape=(16, 1)))


class ExpectationDualBasisOperatorWithPlaneWaveBasisState(unittest.TestCase):

    def setUp(self):
        grid_length = 4
        dimension = 1
        wigner_seitz_radius = 10.
        self.spinless = True
        self.n_spatial_orbitals = grid_length**dimension

        n_qubits = self.n_spatial_orbitals
        self.n_particles = 3

        # Compute appropriate length scale and the corresponding grid.
        length_scale = wigner_seitz_length_scale(wigner_seitz_radius,
                                                 self.n_particles, dimension)

        self.grid1 = Grid(dimension, grid_length, length_scale)
        # Get the occupied orbitals of the plane-wave basis Hartree-Fock state.
        hamiltonian = jellium_model(self.grid1, self.spinless, plane_wave=True)
        hamiltonian = normal_ordered(hamiltonian)
        hamiltonian.compress()

        occupied_states = numpy.array(
            lowest_single_particle_energy_states(hamiltonian, self.n_particles))
        self.hf_state_index1 = numpy.sum(2**occupied_states)

        self.hf_state1 = numpy.zeros(2**n_qubits)
        self.hf_state1[self.hf_state_index1] = 1.0

        self.orbital_occupations1 = [
            digit == '1' for digit in bin(self.hf_state_index1)[2:]
        ][::-1]
        self.occupied_orbitals1 = [
            index for index, occupied in enumerate(self.orbital_occupations1)
            if occupied
        ]

        self.reversed_occupied_orbitals1 = list(self.occupied_orbitals1)
        for i in range(len(self.reversed_occupied_orbitals1)):
            self.reversed_occupied_orbitals1[i] = -1 + int(
                numpy.log2(self.hf_state1.shape[0])
            ) - self.reversed_occupied_orbitals1[i]

        self.reversed_hf_state_index1 = sum(
            2**index for index in self.reversed_occupied_orbitals1)

    def test_1body_hopping_operator_1D(self):
        operator = FermionOperator('2^ 0')
        operator = normal_ordered(operator)
        transformed_operator = normal_ordered(
            fourier_transform(operator, self.grid1, self.spinless))

        expected = expectation(get_sparse_operator(transformed_operator),
                               self.hf_state1)
        actual = expectation_db_operator_with_pw_basis_state(
            operator, self.reversed_occupied_orbitals1, self.n_spatial_orbitals,
            self.grid1, self.spinless)
        self.assertAlmostEqual(expected, actual)

    def test_1body_number_operator_1D(self):
        operator = FermionOperator('2^ 2')
        operator = normal_ordered(operator)
        transformed_operator = normal_ordered(
            fourier_transform(operator, self.grid1, self.spinless))

        expected = expectation(get_sparse_operator(transformed_operator),
                               self.hf_state1)
        actual = expectation_db_operator_with_pw_basis_state(
            operator, self.reversed_occupied_orbitals1, self.n_spatial_orbitals,
            self.grid1, self.spinless)
        self.assertAlmostEqual(expected, actual)

    def test_2body_partial_number_operator_high_1D(self):
        operator = FermionOperator('2^ 1^ 2 0')
        operator = normal_ordered(operator)
        transformed_operator = normal_ordered(
            fourier_transform(operator, self.grid1, self.spinless))

        expected = expectation(get_sparse_operator(transformed_operator),
                               self.hf_state1)
        actual = expectation_db_operator_with_pw_basis_state(
            operator, self.reversed_occupied_orbitals1, self.n_spatial_orbitals,
            self.grid1, self.spinless)
        self.assertAlmostEqual(expected, actual)

    def test_2body_partial_number_operator_mid_1D(self):
        operator = FermionOperator('1^ 0^ 1 2')
        operator = normal_ordered(operator)
        transformed_operator = normal_ordered(
            fourier_transform(operator, self.grid1, self.spinless))

        expected = expectation(get_sparse_operator(transformed_operator),
                               self.hf_state1)
        actual = expectation_db_operator_with_pw_basis_state(
            operator, self.reversed_occupied_orbitals1, self.n_spatial_orbitals,
            self.grid1, self.spinless)
        self.assertAlmostEqual(expected, actual)

    def test_3body_double_number_operator_1D(self):
        operator = FermionOperator('3^ 2^ 1^ 3 1 0')
        operator = normal_ordered(operator)
        transformed_operator = normal_ordered(
            fourier_transform(operator, self.grid1, self.spinless))

        expected = expectation(get_sparse_operator(transformed_operator),
                               self.hf_state1)
        actual = expectation_db_operator_with_pw_basis_state(
            operator, self.reversed_occupied_orbitals1, self.n_spatial_orbitals,
            self.grid1, self.spinless)
        self.assertAlmostEqual(expected, actual)

    def test_2body_adjacent_number_operator_1D(self):
        operator = FermionOperator('3^ 2^ 2 1')
        operator = normal_ordered(operator)
        transformed_operator = normal_ordered(
            fourier_transform(operator, self.grid1, self.spinless))

        expected = expectation(get_sparse_operator(transformed_operator),
                               self.hf_state1)
        actual = expectation_db_operator_with_pw_basis_state(
            operator, self.reversed_occupied_orbitals1, self.n_spatial_orbitals,
            self.grid1, self.spinless)
        self.assertAlmostEqual(expected, actual)

    def test_1d5_with_spin_10particles(self):
        dimension = 1
        grid_length = 5
        n_spatial_orbitals = grid_length**dimension
        wigner_seitz_radius = 9.3

        spinless = False
        n_qubits = n_spatial_orbitals
        if not spinless:
            n_qubits *= 2
        n_particles_big = 10

        length_scale = wigner_seitz_length_scale(wigner_seitz_radius,
                                                 n_particles_big, dimension)

        self.grid3 = Grid(dimension, grid_length, length_scale)
        # Get the occupied orbitals of the plane-wave basis Hartree-Fock state.
        hamiltonian = jellium_model(self.grid3, spinless, plane_wave=True)
        hamiltonian = normal_ordered(hamiltonian)
        hamiltonian.compress()

        occupied_states = numpy.array(
            lowest_single_particle_energy_states(hamiltonian, n_particles_big))
        self.hf_state_index3 = numpy.sum(2**occupied_states)

        self.hf_state3 = csc_matrix(([1.0], ([self.hf_state_index3], [0])),
                                    shape=(2**n_qubits, 1))

        self.orbital_occupations3 = [
            digit == '1' for digit in bin(self.hf_state_index3)[2:]
        ][::-1]
        self.occupied_orbitals3 = [
            index for index, occupied in enumerate(self.orbital_occupations3)
            if occupied
        ]

        self.reversed_occupied_orbitals3 = list(self.occupied_orbitals3)
        for i in range(len(self.reversed_occupied_orbitals3)):
            self.reversed_occupied_orbitals3[i] = -1 + int(
                numpy.log2(self.hf_state3.shape[0])
            ) - self.reversed_occupied_orbitals3[i]

        self.reversed_hf_state_index3 = sum(
            2**index for index in self.reversed_occupied_orbitals3)

        operator = (FermionOperator('6^ 0^ 1^ 3 5 4', 2) +
                    FermionOperator('7^ 6^ 5 4', -3.7j) +
                    FermionOperator('3^ 3', 2.1) + FermionOperator('3^ 2', 1.7))
        operator = normal_ordered(operator)
        normal_ordered(fourier_transform(operator, self.grid3, spinless))

        expected = 2.1
        # Calculated from expectation(get_sparse_operator(
        #    transformed_operator), self.hf_state3)
        actual = expectation_db_operator_with_pw_basis_state(
            operator, self.reversed_occupied_orbitals3, n_spatial_orbitals,
            self.grid3, spinless)

        self.assertAlmostEqual(expected, actual)

    def test_1d5_with_spin_7particles(self):
        dimension = 1
        grid_length = 5
        n_spatial_orbitals = grid_length**dimension
        wigner_seitz_radius = 9.3

        spinless = False
        n_qubits = n_spatial_orbitals
        if not spinless:
            n_qubits *= 2
        n_particles_big = 7

        length_scale = wigner_seitz_length_scale(wigner_seitz_radius,
                                                 n_particles_big, dimension)

        self.grid3 = Grid(dimension, grid_length, length_scale)
        # Get the occupied orbitals of the plane-wave basis Hartree-Fock state.
        hamiltonian = jellium_model(self.grid3, spinless, plane_wave=True)
        hamiltonian = normal_ordered(hamiltonian)
        hamiltonian.compress()

        occupied_states = numpy.array(
            lowest_single_particle_energy_states(hamiltonian, n_particles_big))
        self.hf_state_index3 = numpy.sum(2**occupied_states)

        self.hf_state3 = csc_matrix(([1.0], ([self.hf_state_index3], [0])),
                                    shape=(2**n_qubits, 1))

        self.orbital_occupations3 = [
            digit == '1' for digit in bin(self.hf_state_index3)[2:]
        ][::-1]
        self.occupied_orbitals3 = [
            index for index, occupied in enumerate(self.orbital_occupations3)
            if occupied
        ]

        self.reversed_occupied_orbitals3 = list(self.occupied_orbitals3)
        for i in range(len(self.reversed_occupied_orbitals3)):
            self.reversed_occupied_orbitals3[i] = -1 + int(
                numpy.log2(self.hf_state3.shape[0])
            ) - self.reversed_occupied_orbitals3[i]

        self.reversed_hf_state_index3 = sum(
            2**index for index in self.reversed_occupied_orbitals3)

        operator = (FermionOperator('6^ 0^ 1^ 3 5 4', 2) +
                    FermionOperator('7^ 2^ 4 1') +
                    FermionOperator('3^ 3', 2.1) +
                    FermionOperator('5^ 3^ 1 0', 7.3))
        operator = normal_ordered(operator)
        normal_ordered(fourier_transform(operator, self.grid3, spinless))

        expected = 1.66 - 0.0615536707435j
        # Calculated with expected = expectation(get_sparse_operator(
        #    transformed_operator), self.hf_state3)
        actual = expectation_db_operator_with_pw_basis_state(
            operator, self.reversed_occupied_orbitals3, n_spatial_orbitals,
            self.grid3, spinless)

        self.assertAlmostEqual(expected, actual)

    def test_3d2_spinless(self):
        dimension = 3
        grid_length = 2
        n_spatial_orbitals = grid_length**dimension
        wigner_seitz_radius = 9.3

        spinless = True
        n_qubits = n_spatial_orbitals
        n_particles_big = 5

        length_scale = wigner_seitz_length_scale(wigner_seitz_radius,
                                                 n_particles_big, dimension)

        self.grid3 = Grid(dimension, grid_length, length_scale)
        # Get the occupied orbitals of the plane-wave basis Hartree-Fock state.
        hamiltonian = jellium_model(self.grid3, spinless, plane_wave=True)
        hamiltonian = normal_ordered(hamiltonian)
        hamiltonian.compress()

        occupied_states = numpy.array(
            lowest_single_particle_energy_states(hamiltonian, n_particles_big))
        self.hf_state_index3 = numpy.sum(2**occupied_states)

        self.hf_state3 = csc_matrix(([1.0], ([self.hf_state_index3], [0])),
                                    shape=(2**n_qubits, 1))

        self.orbital_occupations3 = [
            digit == '1' for digit in bin(self.hf_state_index3)[2:]
        ][::-1]
        self.occupied_orbitals3 = [
            index for index, occupied in enumerate(self.orbital_occupations3)
            if occupied
        ]

        self.reversed_occupied_orbitals3 = list(self.occupied_orbitals3)
        for i in range(len(self.reversed_occupied_orbitals3)):
            self.reversed_occupied_orbitals3[i] = -1 + int(
                numpy.log2(self.hf_state3.shape[0])
            ) - self.reversed_occupied_orbitals3[i]

        self.reversed_hf_state_index3 = sum(
            2**index for index in self.reversed_occupied_orbitals3)

        operator = (FermionOperator('4^ 2^ 3^ 5 5 4', 2) +
                    FermionOperator('7^ 6^ 7 4', -3.7j) +
                    FermionOperator('3^ 7', 2.1))
        operator = normal_ordered(operator)
        normal_ordered(fourier_transform(operator, self.grid3, spinless))
        expected = -0.2625 - 0.4625j
        # Calculated with expectation(get_sparse_operator(
        #    transformed_operator), self.hf_state3)
        actual = expectation_db_operator_with_pw_basis_state(
            operator, self.reversed_occupied_orbitals3, n_spatial_orbitals,
            self.grid3, spinless)

        self.assertAlmostEqual(expected, actual)

    def test_3d2_with_spin(self):
        dimension = 3
        grid_length = 2
        n_spatial_orbitals = grid_length**dimension
        wigner_seitz_radius = 9.3

        spinless = False
        n_qubits = n_spatial_orbitals
        if not spinless:
            n_qubits *= 2
        n_particles_big = 9

        length_scale = wigner_seitz_length_scale(wigner_seitz_radius,
                                                 n_particles_big, dimension)

        self.grid3 = Grid(dimension, grid_length, length_scale)
        # Get the occupied orbitals of the plane-wave basis Hartree-Fock state.
        hamiltonian = jellium_model(self.grid3, spinless, plane_wave=True)
        hamiltonian = normal_ordered(hamiltonian)
        hamiltonian.compress()

        occupied_states = numpy.array(
            lowest_single_particle_energy_states(hamiltonian, n_particles_big))
        self.hf_state_index3 = numpy.sum(2**occupied_states)

        self.hf_state3 = csc_matrix(([1.0], ([self.hf_state_index3], [0])),
                                    shape=(2**n_qubits, 1))

        self.orbital_occupations3 = [
            digit == '1' for digit in bin(self.hf_state_index3)[2:]
        ][::-1]
        self.occupied_orbitals3 = [
            index for index, occupied in enumerate(self.orbital_occupations3)
            if occupied
        ]

        self.reversed_occupied_orbitals3 = list(self.occupied_orbitals3)
        for i in range(len(self.reversed_occupied_orbitals3)):
            self.reversed_occupied_orbitals3[i] = -1 + int(
                numpy.log2(self.hf_state3.shape[0])
            ) - self.reversed_occupied_orbitals3[i]

        self.reversed_hf_state_index3 = sum(
            2**index for index in self.reversed_occupied_orbitals3)

        operator = (FermionOperator('4^ 2^ 3^ 5 5 4', 2) +
                    FermionOperator('7^ 6^ 7 4', -3.7j) +
                    FermionOperator('3^ 7', 2.1))
        operator = normal_ordered(operator)
        normal_ordered(fourier_transform(operator, self.grid3, spinless))

        expected = -0.2625 - 0.578125j
        # Calculated from expected = expectation(get_sparse_operator(
        #    transformed_operator), self.hf_state3)
        actual = expectation_db_operator_with_pw_basis_state(
            operator, self.reversed_occupied_orbitals3, n_spatial_orbitals,
            self.grid3, spinless)

        self.assertAlmostEqual(expected, actual)


class GetGapTest(unittest.TestCase):

    def test_get_gap(self):
        operator = QubitOperator('Y0 X1') + QubitOperator('Z0 Z1')
        self.assertAlmostEqual(get_gap(get_sparse_operator(operator)), 2.0)

    def test_get_gap_nonhermitian_error(self):
        operator = (QubitOperator('X0 Y1', 1 + 1j) +
                    QubitOperator('Z0 Z1', 1j) + QubitOperator((), 2 + 1j))
        with self.assertRaises(ValueError):
            get_gap(get_sparse_operator(operator))


class InnerProductTest(unittest.TestCase):

    def test_inner_product(self):
        state_1 = numpy.array([1., 1.j])
        state_2 = numpy.array([1., -1.j])

        self.assertAlmostEqual(inner_product(state_1, state_1), 2.)
        self.assertAlmostEqual(inner_product(state_1, state_2), 0.)


class BosonSparseTest(unittest.TestCase):

    def setUp(self):
        self.hbar = 1.
        self.d = 5
        self.b = numpy.diag(numpy.sqrt(numpy.arange(1, self.d)), 1)
        self.bd = self.b.conj().T
        self.q = numpy.sqrt(self.hbar / 2) * (self.b + self.bd)
        self.p = -1j * numpy.sqrt(self.hbar / 2) * (self.b - self.bd)
        self.Id = numpy.identity(self.d)

    def test_boson_ladder_noninteger_trunc(self):
        with self.assertRaises(ValueError):
            boson_ladder_sparse(1, 0, 0, 0.1)

        with self.assertRaises(ValueError):
            boson_ladder_sparse(1, 0, 0, -1)

        with self.assertRaises(ValueError):
            boson_ladder_sparse(1, 0, 0, 0)

    def test_boson_ladder_destroy_one_mode(self):
        b = boson_ladder_sparse(1, 0, 0, self.d).toarray()
        self.assertTrue(numpy.allclose(b, self.b))

    def test_boson_ladder_create_one_mode(self):
        bd = boson_ladder_sparse(1, 0, 1, self.d).toarray()
        self.assertTrue(numpy.allclose(bd, self.bd))

    def test_boson_ladder_single_adjoint(self):
        b = boson_ladder_sparse(1, 0, 0, self.d).toarray()
        bd = boson_ladder_sparse(1, 0, 1, self.d).toarray()
        self.assertTrue(numpy.allclose(b.conj().T, bd))

    def test_boson_ladder_two_mode(self):
        res = boson_ladder_sparse(2, 0, 0, self.d).toarray()
        expected = numpy.kron(self.b, self.Id)
        self.assertTrue(numpy.allclose(res, expected))

        res = boson_ladder_sparse(2, 1, 0, self.d).toarray()
        expected = numpy.kron(self.Id, self.b)
        self.assertTrue(numpy.allclose(res, expected))

    def test_single_quad_noninteger_trunc(self):
        with self.assertRaises(ValueError):
            single_quad_op_sparse(1, 0, 'q', self.hbar, 0.1)

        with self.assertRaises(ValueError):
            single_quad_op_sparse(1, 0, 'q', self.hbar, -1)

        with self.assertRaises(ValueError):
            single_quad_op_sparse(1, 0, 'q', self.hbar, 0)

    def test_single_quad_q_one_mode(self):
        res = single_quad_op_sparse(1, 0, 'q', self.hbar, self.d).toarray()
        self.assertTrue(numpy.allclose(res, self.q))
        self.assertTrue(numpy.allclose(res, res.conj().T))

    def test_single_quad_p_one_mode(self):
        res = single_quad_op_sparse(1, 0, 'p', self.hbar, self.d).toarray()
        self.assertTrue(numpy.allclose(res, self.p))
        self.assertTrue(numpy.allclose(res, res.conj().T))

    def test_single_quad_two_mode(self):
        res = single_quad_op_sparse(2, 0, 'q', self.hbar, self.d).toarray()
        expected = numpy.kron(self.q, self.Id)
        self.assertTrue(numpy.allclose(res, expected))

        res = single_quad_op_sparse(2, 1, 'p', self.hbar, self.d).toarray()
        expected = numpy.kron(self.Id, self.p)
        self.assertTrue(numpy.allclose(res, expected))

    def test_boson_operator_sparse_trunc(self):
        op = BosonOperator('0')
        with self.assertRaises(ValueError):
            boson_operator_sparse(op, 0.1)

        with self.assertRaises(ValueError):
            boson_operator_sparse(op, -1)

        with self.assertRaises(ValueError):
            boson_operator_sparse(op, 0)

    def test_boson_operator_invalid_op(self):
        op = FermionOperator('0')
        with self.assertRaises(ValueError):
            boson_operator_sparse(op, self.d)

    def test_boson_operator_sparse_empty(self):
        for op in (BosonOperator(), QuadOperator()):
            res = boson_operator_sparse(op, self.d)
            self.assertEqual(res, numpy.array([[0]]))

    def test_boson_operator_sparse_identity(self):
        for op in (BosonOperator(''), QuadOperator('')):
            res = boson_operator_sparse(op, self.d)
            self.assertEqual(res, numpy.array([[1]]))

    def test_boson_operator_sparse_single(self):
        op = BosonOperator('0')
        res = boson_operator_sparse(op, self.d).toarray()
        self.assertTrue(numpy.allclose(res, self.b))

        op = BosonOperator('0^')
        res = boson_operator_sparse(op, self.d).toarray()
        self.assertTrue(numpy.allclose(res, self.bd))

        op = QuadOperator('q0')
        res = boson_operator_sparse(op, self.d, self.hbar).toarray()
        self.assertTrue(numpy.allclose(res, self.q))

        op = QuadOperator('p0')
        res = boson_operator_sparse(op, self.d, self.hbar).toarray()
        self.assertTrue(numpy.allclose(res, self.p))

    def test_boson_operator_sparse_number(self):
        op = BosonOperator('0^ 0')
        res = boson_operator_sparse(op, self.d).toarray()
        self.assertTrue(numpy.allclose(res, numpy.dot(self.bd, self.b)))

    def test_boson_operator_sparse_multi_mode(self):
        op = BosonOperator('0^ 1 1^ 2')
        res = boson_operator_sparse(op, self.d).toarray()

        b0 = boson_ladder_sparse(3, 0, 0, self.d).toarray()
        b1 = boson_ladder_sparse(3, 1, 0, self.d).toarray()
        b2 = boson_ladder_sparse(3, 2, 0, self.d).toarray()

        expected = multi_dot([b0.T, b1, b1.T, b2])
        self.assertTrue(numpy.allclose(res, expected))

        op = QuadOperator('q0 p0 p1')
        res = boson_operator_sparse(op, self.d, self.hbar).toarray()

        expected = numpy.identity(self.d**2)
        for term in op.terms:
            for i, j in term:
                expected = expected.dot(
                    single_quad_op_sparse(2, i, j, self.hbar, self.d).toarray())
        self.assertTrue(numpy.allclose(res, expected))

    def test_boson_operator_sparse_addition(self):
        op = BosonOperator('0^ 1')
        op += BosonOperator('0 0^')
        res = boson_operator_sparse(op, self.d).toarray()

        b0 = boson_ladder_sparse(2, 0, 0, self.d).toarray()
        b1 = boson_ladder_sparse(2, 1, 0, self.d).toarray()

        expected = numpy.dot(b0.T, b1) + numpy.dot(b0, b0.T)
        self.assertTrue(numpy.allclose(res, expected))


class GetNumberPreservingSparseOperatorIntegrationTestLiH(unittest.TestCase):

    def setUp(self):
        # Set up molecule.
        geometry = [('Li', (0., 0., 0.)), ('H', (0., 0., 1.45))]
        basis = 'sto-3g'
        multiplicity = 1
        filename = os.path.join(DATA_DIRECTORY, 'H1-Li1_sto-3g_singlet_1.45')
        self.molecule = MolecularData(geometry,
                                      basis,
                                      multiplicity,
                                      filename=filename)
        self.molecule.load()

        # Get molecular Hamiltonian.
        self.molecular_hamiltonian = self.molecule.get_molecular_hamiltonian()

        self.hubbard_hamiltonian = fermi_hubbard(2,
                                                 2,
                                                 1.0,
                                                 4.0,
                                                 chemical_potential=.2,
                                                 magnetic_field=0.0,
                                                 spinless=False)

    def test_exceptions(self):
        op = FermionOperator('1')
        with self.assertRaises(ValueError):
            _ = get_number_preserving_sparse_operator(op, 2, 1)

    def test_number_on_reference(self):
        sum_n_op = FermionOperator()
        sum_sparse_n_op = get_number_preserving_sparse_operator(
            sum_n_op,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False)

        space_size = sum_sparse_n_op.shape[0]
        reference = numpy.zeros((space_size))
        reference[0] = 1.0

        for i in range(self.molecule.n_qubits):
            n_op = FermionOperator(((i, 1), (i, 0)))
            sum_n_op += n_op

            sparse_n_op = get_number_preserving_sparse_operator(
                n_op,
                self.molecule.n_qubits,
                self.molecule.n_electrons,
                spin_preserving=False)

            sum_sparse_n_op += sparse_n_op

            expectation = reference.dot(sparse_n_op.dot(reference))

            if i < self.molecule.n_electrons:
                assert expectation == 1.0
            else:
                assert expectation == 0.0

        convert_after_adding = get_number_preserving_sparse_operator(
            sum_n_op,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False)

        assert scipy.sparse.linalg.norm(convert_after_adding -
                                        sum_sparse_n_op) < 1E-9

        assert reference.dot(sum_sparse_n_op.dot(reference)) - \
            self.molecule.n_electrons < 1E-9

    def test_space_size_correct(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True)

        space_size = sparse_ham.shape[0]

        # Naive Hilbert space size is 2**12, or 4096.
        assert space_size == 225

    def test_hf_energy(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True)

        space_size = sparse_ham.shape[0]
        reference = numpy.zeros((space_size))
        reference[0] = 1.0

        sparse_hf_energy = reference.dot(sparse_ham.dot(reference))

        assert numpy.linalg.norm(sparse_hf_energy -
                                 self.molecule.hf_energy) < 1E-9

    def test_one_body_hf_energy(self):
        one_body_part = self.molecular_hamiltonian
        one_body_part.two_body_tensor = numpy.zeros_like(
            one_body_part.two_body_tensor)

        one_body_fop = get_fermion_operator(one_body_part)
        one_body_regular_sparse_op = get_sparse_operator(one_body_fop)

        make_hf_fop = FermionOperator(((3, 1), (2, 1), (1, 1), (0, 1)))
        make_hf_sparse_op = get_sparse_operator(make_hf_fop, n_qubits=12)

        hf_state = numpy.zeros((2**12))
        hf_state[0] = 1.0
        hf_state = make_hf_sparse_op.dot(hf_state)

        regular_sparse_hf_energy = \
            (hf_state.dot(one_body_regular_sparse_op.dot(hf_state))).real

        one_body_sparse_op = get_number_preserving_sparse_operator(
            one_body_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True)

        space_size = one_body_sparse_op.shape[0]
        reference = numpy.zeros((space_size))
        reference[0] = 1.0

        sparse_hf_energy = reference.dot(one_body_sparse_op.dot(reference))

        assert numpy.linalg.norm(sparse_hf_energy -
                                 regular_sparse_hf_energy) < 1E-9

    def test_ground_state_energy(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True)

        eig_val, _ = scipy.sparse.linalg.eigsh(sparse_ham, k=1, which='SA')

        assert numpy.abs(eig_val[0] - self.molecule.fci_energy) < 1E-9

    def test_doubles_are_subset(self):
        reference_determinants = [[
            True, True, True, True, False, False, False, False, False, False,
            False, False
        ],
                                  [
                                      True, True, False, False, False, False,
                                      True, True, False, False, False, False
                                  ]]

        for reference_determinant in reference_determinants:
            reference_determinant = numpy.asarray(reference_determinant)
            doubles_state_array = numpy.asarray(
                list(
                    _iterate_basis_(reference_determinant,
                                    excitation_level=2,
                                    spin_preserving=True)))
            doubles_int_state_array = doubles_state_array.dot(
                1 << numpy.arange(doubles_state_array.shape[1])[::-1])

            all_state_array = numpy.asarray(
                list(
                    _iterate_basis_(reference_determinant,
                                    excitation_level=4,
                                    spin_preserving=True)))
            all_int_state_array = all_state_array.dot(
                1 << numpy.arange(all_state_array.shape[1])[::-1])

            for item in doubles_int_state_array:
                assert item in all_int_state_array

        for reference_determinant in reference_determinants:
            reference_determinant = numpy.asarray(reference_determinant)
            doubles_state_array = numpy.asarray(
                list(
                    _iterate_basis_(reference_determinant,
                                    excitation_level=2,
                                    spin_preserving=True)))
            doubles_int_state_array = doubles_state_array.dot(
                1 << numpy.arange(doubles_state_array.shape[1])[::-1])

            all_state_array = numpy.asarray(
                list(
                    _iterate_basis_(reference_determinant,
                                    excitation_level=4,
                                    spin_preserving=False)))
            all_int_state_array = all_state_array.dot(
                1 << numpy.arange(all_state_array.shape[1])[::-1])

            for item in doubles_int_state_array:
                assert item in all_int_state_array

        for reference_determinant in reference_determinants:
            reference_determinant = numpy.asarray(reference_determinant)
            doubles_state_array = numpy.asarray(
                list(
                    _iterate_basis_(reference_determinant,
                                    excitation_level=2,
                                    spin_preserving=False)))
            doubles_int_state_array = doubles_state_array.dot(
                1 << numpy.arange(doubles_state_array.shape[1])[::-1])

            all_state_array = numpy.asarray(
                list(
                    _iterate_basis_(reference_determinant,
                                    excitation_level=4,
                                    spin_preserving=False)))
            all_int_state_array = all_state_array.dot(
                1 << numpy.arange(all_state_array.shape[1])[::-1])

            for item in doubles_int_state_array:
                assert item in all_int_state_array

    def test_full_ham_hermitian(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True)

        assert scipy.sparse.linalg.norm(sparse_ham - sparse_ham.getH()) < 1E-9

    def test_full_ham_hermitian_non_spin_preserving(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False)

        assert scipy.sparse.linalg.norm(sparse_ham - sparse_ham.getH()) < 1E-9

    def test_singles_simple_one_body_term_hermitian(self):
        fop = FermionOperator(((3, 1), (1, 0)))
        fop_conj = FermionOperator(((1, 1), (3, 0)))

        sparse_op = get_number_preserving_sparse_operator(
            fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1)

        sparse_op_conj = get_number_preserving_sparse_operator(
            fop_conj,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1)

        assert scipy.sparse.linalg.norm(sparse_op -
                                        sparse_op_conj.getH()) < 1E-9

    def test_singles_simple_two_body_term_hermitian(self):
        fop = FermionOperator(((3, 1), (8, 1), (1, 0), (4, 0)))
        fop_conj = FermionOperator(((4, 1), (1, 1), (8, 0), (3, 0)))

        sparse_op = get_number_preserving_sparse_operator(
            fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1)

        sparse_op_conj = get_number_preserving_sparse_operator(
            fop_conj,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1)

        assert scipy.sparse.linalg.norm(sparse_op -
                                        sparse_op_conj.getH()) < 1E-9

    def test_singles_repeating_two_body_term_hermitian(self):
        fop = FermionOperator(((3, 1), (1, 1), (5, 0), (1, 0)))
        fop_conj = FermionOperator(((5, 1), (1, 1), (3, 0), (1, 0)))

        sparse_op = get_number_preserving_sparse_operator(
            fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1)

        sparse_op_conj = get_number_preserving_sparse_operator(
            fop_conj,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1)

        assert scipy.sparse.linalg.norm(sparse_op -
                                        sparse_op_conj.getH()) < 1E-9

    def test_singles_ham_hermitian(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=1)

        assert scipy.sparse.linalg.norm(sparse_ham - sparse_ham.getH()) < 1E-9

    def test_singles_ham_hermitian_non_spin_preserving(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False,
            excitation_level=1)

        assert scipy.sparse.linalg.norm(sparse_ham - sparse_ham.getH()) < 1E-9

    def test_cisd_energy(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True,
            excitation_level=2)

        eig_val, _ = scipy.sparse.linalg.eigsh(sparse_ham, k=1, which='SA')

        assert numpy.abs(eig_val[0] - self.molecule.cisd_energy) < 1E-9

    def test_cisd_energy_non_spin_preserving(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False,
            excitation_level=2)

        eig_val, _ = scipy.sparse.linalg.eigsh(sparse_ham, k=1, which='SA')

        assert numpy.abs(eig_val[0] - self.molecule.cisd_energy) < 1E-9

    def test_cisd_matches_fci_energy_two_electron_hubbard(self):
        hamiltonian_fop = self.hubbard_hamiltonian

        sparse_ham_cisd = get_number_preserving_sparse_operator(
            hamiltonian_fop, 8, 2, spin_preserving=True, excitation_level=2)

        sparse_ham_fci = get_sparse_operator(hamiltonian_fop, n_qubits=8)

        eig_val_cisd, _ = scipy.sparse.linalg.eigsh(sparse_ham_cisd,
                                                    k=1,
                                                    which='SA')
        eig_val_fci, _ = scipy.sparse.linalg.eigsh(sparse_ham_fci,
                                                   k=1,
                                                   which='SA')

        assert numpy.abs(eig_val_cisd[0] - eig_val_fci[0]) < 1E-9

    def test_weird_determinant_matches_fci_energy_two_electron_hubbard(self):
        hamiltonian_fop = self.hubbard_hamiltonian

        sparse_ham_cisd = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            8,
            2,
            spin_preserving=True,
            excitation_level=2,
            reference_determinant=numpy.asarray(
                [False, False, True, True, False, False, False, False]))

        sparse_ham_fci = get_sparse_operator(hamiltonian_fop, n_qubits=8)

        eig_val_cisd, _ = scipy.sparse.linalg.eigsh(sparse_ham_cisd,
                                                    k=1,
                                                    which='SA')
        eig_val_fci, _ = scipy.sparse.linalg.eigsh(sparse_ham_fci,
                                                   k=1,
                                                   which='SA')

        assert numpy.abs(eig_val_cisd[0] - eig_val_fci[0]) < 1E-9

    def test_number_restricted_spectra_match_molecule(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham_number_preserving = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=False)

        sparse_ham = get_sparse_operator(hamiltonian_fop,
                                         self.molecule.n_qubits)

        sparse_ham_restricted_number_preserving = jw_number_restrict_operator(
            sparse_ham,
            n_electrons=self.molecule.n_electrons,
            n_qubits=self.molecule.n_qubits)

        spectrum_from_new_sparse_method = sparse_eigenspectrum(
            sparse_ham_number_preserving)

        spectrum_from_old_sparse_method = sparse_eigenspectrum(
            sparse_ham_restricted_number_preserving)

        spectral_deviation = numpy.amax(
            numpy.absolute(spectrum_from_new_sparse_method -
                           spectrum_from_old_sparse_method))
        self.assertAlmostEqual(spectral_deviation, 0.)

    def test_number_restricted_spectra_match_hubbard(self):
        hamiltonian_fop = self.hubbard_hamiltonian

        sparse_ham_number_preserving = get_number_preserving_sparse_operator(
            hamiltonian_fop, 8, 4, spin_preserving=False)

        sparse_ham = get_sparse_operator(hamiltonian_fop, 8)

        sparse_ham_restricted_number_preserving = jw_number_restrict_operator(
            sparse_ham, n_electrons=4, n_qubits=8)

        spectrum_from_new_sparse_method = sparse_eigenspectrum(
            sparse_ham_number_preserving)

        spectrum_from_old_sparse_method = sparse_eigenspectrum(
            sparse_ham_restricted_number_preserving)

        spectral_deviation = numpy.amax(
            numpy.absolute(spectrum_from_new_sparse_method -
                           spectrum_from_old_sparse_method))
        self.assertAlmostEqual(spectral_deviation, 0.)

    def test_number_sz_restricted_spectra_match_molecule(self):
        hamiltonian_fop = get_fermion_operator(self.molecular_hamiltonian)

        sparse_ham_number_sz_preserving = get_number_preserving_sparse_operator(
            hamiltonian_fop,
            self.molecule.n_qubits,
            self.molecule.n_electrons,
            spin_preserving=True)

        sparse_ham = get_sparse_operator(hamiltonian_fop,
                                         self.molecule.n_qubits)

        sparse_ham_restricted_number_sz_preserving = jw_sz_restrict_operator(
            sparse_ham,
            0,
            n_electrons=self.molecule.n_electrons,
            n_qubits=self.molecule.n_qubits)

        spectrum_from_new_sparse_method = sparse_eigenspectrum(
            sparse_ham_number_sz_preserving)

        spectrum_from_old_sparse_method = sparse_eigenspectrum(
            sparse_ham_restricted_number_sz_preserving)

        spectral_deviation = numpy.amax(
            numpy.absolute(spectrum_from_new_sparse_method -
                           spectrum_from_old_sparse_method))
        self.assertAlmostEqual(spectral_deviation, 0.)

    def test_number_sz_restricted_spectra_match_hubbard(self):
        hamiltonian_fop = self.hubbard_hamiltonian

        sparse_ham_number_sz_preserving = get_number_preserving_sparse_operator(
            hamiltonian_fop, 8, 4, spin_preserving=True)

        sparse_ham = get_sparse_operator(hamiltonian_fop, 8)

        sparse_ham_restricted_number_sz_preserving = jw_sz_restrict_operator(
            sparse_ham, 0, n_electrons=4, n_qubits=8)

        spectrum_from_new_sparse_method = sparse_eigenspectrum(
            sparse_ham_number_sz_preserving)

        spectrum_from_old_sparse_method = sparse_eigenspectrum(
            sparse_ham_restricted_number_sz_preserving)

        spectral_deviation = numpy.amax(
            numpy.absolute(spectrum_from_new_sparse_method -
                           spectrum_from_old_sparse_method))
        self.assertAlmostEqual(spectral_deviation, 0.)


class GetSparseOperatorQubitTest(unittest.TestCase):

    def test_sparse_matrix_Y(self):
        term = QubitOperator(((0, 'Y'),))
        sparse_operator = get_sparse_operator(term)
        self.assertEqual(list(sparse_operator.data), [1j, -1j])
        self.assertEqual(list(sparse_operator.indices), [1, 0])
        self.assertTrue(is_hermitian(sparse_operator))

    def test_sparse_matrix_ZX(self):
        coefficient = 2.
        operators = ((0, 'Z'), (1, 'X'))
        term = QubitOperator(operators, coefficient)
        sparse_operator = get_sparse_operator(term)
        self.assertEqual(list(sparse_operator.data), [2., 2., -2., -2.])
        self.assertEqual(list(sparse_operator.indices), [1, 0, 3, 2])
        self.assertTrue(is_hermitian(sparse_operator))

    def test_sparse_matrix_ZIZ(self):
        operators = ((0, 'Z'), (2, 'Z'))
        term = QubitOperator(operators)
        sparse_operator = get_sparse_operator(term)
        self.assertEqual(list(sparse_operator.data),
                         [1, -1, 1, -1, -1, 1, -1, 1])
        self.assertEqual(list(sparse_operator.indices), list(range(8)))
        self.assertTrue(is_hermitian(sparse_operator))

    def test_sparse_matrix_combo(self):
        qop = (QubitOperator(((0, 'Y'), (1, 'X')), -0.1j) + QubitOperator(
            ((0, 'X'), (1, 'Z')), 3. + 2.j))
        sparse_operator = get_sparse_operator(qop)

        self.assertEqual(
            list(sparse_operator.data),
            [3 + 2j, 0.1, 0.1, -3 - 2j, 3 + 2j, -0.1, -0.1, -3 - 2j])
        self.assertEqual(list(sparse_operator.indices),
                         [2, 3, 2, 3, 0, 1, 0, 1])

    def test_sparse_matrix_zero_1qubit(self):
        sparse_operator = get_sparse_operator(QubitOperator((), 0.0), 1)
        sparse_operator.eliminate_zeros()
        self.assertEqual(len(list(sparse_operator.data)), 0)
        self.assertEqual(sparse_operator.shape, (2, 2))

    def test_sparse_matrix_zero_5qubit(self):
        sparse_operator = get_sparse_operator(QubitOperator((), 0.0), 5)
        sparse_operator.eliminate_zeros()
        self.assertEqual(len(list(sparse_operator.data)), 0)
        self.assertEqual(sparse_operator.shape, (32, 32))

    def test_sparse_matrix_identity_1qubit(self):
        sparse_operator = get_sparse_operator(QubitOperator(()), 1)
        self.assertEqual(list(sparse_operator.data), [1] * 2)
        self.assertEqual(sparse_operator.shape, (2, 2))

    def test_sparse_matrix_identity_5qubit(self):
        sparse_operator = get_sparse_operator(QubitOperator(()), 5)
        self.assertEqual(list(sparse_operator.data), [1] * 32)
        self.assertEqual(sparse_operator.shape, (32, 32))

    def test_sparse_matrix_linearity(self):
        identity = QubitOperator(())
        zzzz = QubitOperator(tuple((i, 'Z') for i in range(4)), 1.0)

        sparse1 = get_sparse_operator(identity + zzzz)
        sparse2 = get_sparse_operator(identity, 4) + get_sparse_operator(zzzz)

        self.assertEqual(list(sparse1.data), [2] * 8)
        self.assertEqual(list(sparse1.indices), [0, 3, 5, 6, 9, 10, 12, 15])
        self.assertEqual(list(sparse2.data), [2] * 8)
        self.assertEqual(list(sparse2.indices), [0, 3, 5, 6, 9, 10, 12, 15])


class GetSparseOperatorFermionTest(unittest.TestCase):

    def test_sparse_matrix_zero_n_qubit(self):
        sparse_operator = get_sparse_operator(FermionOperator.zero(), 4)
        sparse_operator.eliminate_zeros()
        self.assertEqual(len(list(sparse_operator.data)), 0)
        self.assertEqual(sparse_operator.shape, (16, 16))


class GetSparseOperatorBosonTest(unittest.TestCase):

    def setUp(self):
        self.hbar = 1.
        self.d = 4
        self.b = numpy.diag(numpy.sqrt(numpy.arange(1, self.d)), 1)
        self.bd = self.b.conj().T
        self.q = numpy.sqrt(self.hbar / 2) * (self.b + self.bd)

    def test_sparse_matrix_ladder(self):
        sparse_operator = get_sparse_operator(BosonOperator('0'), trunc=self.d)
        self.assertTrue(numpy.allclose(sparse_operator.toarray(), self.b))
        self.assertEqual(sparse_operator.shape, (self.d, self.d))

    def test_sparse_matrix_quad(self):
        sparse_operator = get_sparse_operator(QuadOperator('q0'), trunc=self.d)
        self.assertTrue(numpy.allclose(sparse_operator.toarray(), self.q))
        self.assertEqual(sparse_operator.shape, (self.d, self.d))

    def test_sparse_matrix_error(self):
        with self.assertRaises(TypeError):
            _ = get_sparse_operator(1)


class GetSparseOperatorDiagonalCoulombHamiltonianTest(unittest.TestCase):

    def test_diagonal_coulomb_hamiltonian(self):
        n_qubits = 5
        one_body = random_hermitian_matrix(n_qubits, real=False)
        two_body = random_hermitian_matrix(n_qubits, real=True)
        constant = numpy.random.randn()
        op = DiagonalCoulombHamiltonian(one_body, two_body, constant)

        op1 = get_sparse_operator(op)
        op2 = get_sparse_operator(jordan_wigner(get_fermion_operator(op)))
        diff = op1 - op2
        discrepancy = 0.
        if diff.nnz:
            discrepancy = max(abs(diff.data))
        self.assertAlmostEqual(discrepancy, 0.)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Module to compute Trotter errors in the plane-wave dual basis."""
import numpy

from openfermion.ops.operators import FermionOperator
from openfermion.utils import count_qubits
from openfermion.transforms.opconversions import normal_ordered
from openfermion.utils.commutators import (
    double_commutator, trivially_double_commutes_dual_basis,
    trivially_double_commutes_dual_basis_using_term_info)


def low_depth_second_order_trotter_error_operator(terms,
                                                  indices=None,
                                                  is_hopping_operator=None,
                                                  jellium_only=False,
                                                  verbose=False):
    """Determine the difference between the exact generator of unitary
    evolution and the approximate generator given by the second-order
    Trotter-Suzuki expansion.

    Args:
        terms: a list of FermionOperators in the Hamiltonian in the
               order in which they will be simulated.
        indices: a set of indices the terms act on in the same order as terms.
        is_hopping_operator: a list of whether each term is a hopping operator.
        jellium_only: Whether the terms are from the jellium Hamiltonian only,
                      rather than the full dual basis Hamiltonian (i.e. whether
                      c_i = c for all number operators i^ i, or whether they
                      depend on i as is possible in the general case).
        verbose: Whether to print percentage progress.

    Returns:
        The difference between the true and effective generators of time
            evolution for a single Trotter step.

    Notes: follows Equation 9 of Poulin et al.'s work in "The Trotter Step
        Size Required for Accurate Quantum Simulation of Quantum Chemistry",
        applied to the "stagger"-based Trotter step for detailed in
        Kivlichan et al., "Quantum Simulation of Electronic Structure with
        Linear Depth and Connectivity", arxiv:1711.04789.
    """
    more_info = bool(indices)
    n_terms = len(terms)

    if verbose:
        import time
        start = time.time()

    error_operator = FermionOperator.zero()
    for beta in range(n_terms):
        if verbose and beta % (n_terms // 30) == 0:
            print('%4.3f percent done in' % ((float(beta) / n_terms)**3 * 100),
                  time.time() - start)

        for alpha in range(beta + 1):
            for alpha_prime in range(beta):
                # If we have pre-computed info on indices, use it to determine
                # trivial double commutation.
                if more_info:
                    if (not trivially_double_commutes_dual_basis_using_term_info(  # pylint: disable=C
                            indices[alpha], indices[beta], indices[alpha_prime],
                            is_hopping_operator[alpha],
                            is_hopping_operator[beta],
                            is_hopping_operator[alpha_prime], jellium_only)):
                        # Determine the result of the double commutator.
                        double_com = double_commutator(
                            terms[alpha], terms[beta], terms[alpha_prime],
                            indices[beta], indices[alpha_prime],
                            is_hopping_operator[beta],
                            is_hopping_operator[alpha_prime])
                        if alpha == beta:
                            double_com /= 2.0

                        error_operator += double_com

                # If we don't have more info, check for trivial double
                # commutation using the terms directly.
                elif not trivially_double_commutes_dual_basis(
                        terms[alpha], terms[beta], terms[alpha_prime]):
                    double_com = double_commutator(terms[alpha], terms[beta],
                                                   terms[alpha_prime])

                    if alpha == beta:
                        double_com /= 2.0

                    error_operator += double_com

    error_operator /= 12.0
    return error_operator


def low_depth_second_order_trotter_error_bound(terms,
                                               indices=None,
                                               is_hopping_operator=None,
                                               jellium_only=False,
                                               verbose=False):
    """Numerically upper bound the error in the ground state energy
    for the second-order Trotter-Suzuki expansion.

    Args:
        terms: a list of single-term FermionOperators in the Hamiltonian
            to be simulated.
        indices: a set of indices the terms act on in the same order as terms.
        is_hopping_operator: a list of whether each term is a hopping operator.
        jellium_only: Whether the terms are from the jellium Hamiltonian only,
                      rather than the full dual basis Hamiltonian (i.e. whether
                      c_i = c for all number operators i^ i, or whether they
                      depend on i as is possible in the general case).
        verbose: Whether to print percentage progress.

    Returns:
        A float upper bound on norm of error in the ground state energy.

    Notes:
        Follows Equation 9 of Poulin et al.'s work in "The Trotter Step
        Size Required for Accurate Quantum Simulation of Quantum
        Chemistry" to calculate the error operator, for the "stagger"-based
        Trotter step for detailed in Kivlichan et al., "Quantum Simulation
        of Electronic Structure with Linear Depth and Connectivity",
        arxiv:1711.04789.
    """
    # Return the 1-norm of the error operator (upper bound on error).
    return numpy.sum(
        numpy.absolute(
            list(
                low_depth_second_order_trotter_error_operator(
                    terms, indices, is_hopping_operator, jellium_only,
                    verbose).terms.values())))


def simulation_ordered_grouped_low_depth_terms_with_info(
        hamiltonian, input_ordering=None, external_potential_at_end=False):
    """Give terms from the dual basis Hamiltonian in simulated order.

    Uses the simulation ordering, grouping terms into hopping
    (i^ j + j^ i) and number (i^j^ i j + c_i i^ i + c_j j^ j) operators.
    Pre-computes term information (indices each operator acts on, as
    well as whether each operator is a hopping operator.

    Args:
        hamiltonian (FermionOperator): The Hamiltonian.
        input_ordering (list): The initial Jordan-Wigner canonical order.
                               If no input ordering is specified, defaults to
                               [0..n_qubits] where n_qubits is the number of
                               qubits in the Hamiltonian.
        external_potential_at_end (bool): Whether to include the rotations from
            the external potential at the end of the Trotter step, or
            intersperse them throughout it.

    Returns:
        A 3-tuple of terms from the Hamiltonian in order of simulation,
        the indices they act on, and whether they are hopping operators
        (both also in the same order).

    Notes:
        Follows the "stagger"-based simulation order discussed in Kivlichan
        et al., "Quantum Simulation of Electronic Structure with Linear
        Depth and Connectivity", arxiv:1711.04789; as such, the only
        permitted types of terms are hopping (i^ j + j^ i) and potential
        terms which are products of at most two number operators.
    """
    n_qubits = count_qubits(hamiltonian)
    hamiltonian = normal_ordered(hamiltonian)

    ordered_terms = []
    ordered_indices = []
    ordered_is_hopping_operator = []

    # If no input mode ordering is specified, default to range(n_qubits).
    try:
        input_ordering = list(input_ordering)
    except TypeError:
        input_ordering = list(range(n_qubits))

    # Half a second-order Trotter step reverses the input ordering: this tells
    # us how much we need to include in the ordered list of terms.
    final_ordering = list(reversed(input_ordering))

    # Follow odd-even transposition sort. In alternating steps, swap each even
    # qubits with the odd qubit to its right, and in the next step swap each
    # the odd qubits with the even qubit to its right. Do this until the input
    # ordering has been reversed.
    parity = 0
    while input_ordering != final_ordering:
        results = stagger_with_info(hamiltonian, input_ordering, parity,
                                    external_potential_at_end)
        terms_in_layer, indices_in_layer, is_hopping_operator_in_layer = (
            results)

        ordered_terms.extend(terms_in_layer)
        ordered_indices.extend(indices_in_layer)
        ordered_is_hopping_operator.extend(is_hopping_operator_in_layer)

        # Alternate even and odd steps of the reversal procedure.
        parity = 1 - parity

    # If all the rotations from the external potential are in the final layer,
    # i.e. we don't intersperse them throughout the Trotter step.
    if external_potential_at_end:
        terms_in_final_layer = []
        indices_in_final_layer = []
        is_hopping_operator_in_final_layer = []

        for qubit in range(n_qubits):
            coeff = hamiltonian.terms.get(((qubit, 1), (qubit, 0)), 0.0)
            if coeff:
                terms_in_final_layer.append(
                    FermionOperator(((qubit, 1), (qubit, 0)), coeff))
                indices_in_final_layer.append(set((qubit,)))
                is_hopping_operator_in_final_layer.append(False)

        ordered_terms.extend(terms_in_final_layer)
        ordered_indices.extend(indices_in_final_layer)
        ordered_is_hopping_operator.extend(is_hopping_operator_in_final_layer)

    return (ordered_terms, ordered_indices, ordered_is_hopping_operator)


def stagger_with_info(hamiltonian,
                      input_ordering,
                      parity,
                      external_potential_at_end=False):
    """Give terms simulated in a single stagger of a Trotter step.

    Groups terms into hopping (i^ j + j^ i) and number
    (i^j^ i j + c_i i^ i + c_j j^ j) operators.
    Pre-computes term information (indices each operator acts on, as
    well as whether each operator is a hopping operator).

    Args:
        hamiltonian (FermionOperator): The Hamiltonian.
        input_ordering (list): The initial Jordan-Wigner canonical order.
        parity (boolean): Whether to determine the terms from the next even
            (False = 0) or odd (True = 1) stagger.
        external_potential_at_end (bool): Whether to include the rotations from
            the external potential at the end of the Trotter step, or
            intersperse them throughout it.

    Returns:
        A 3-tuple of terms from the Hamiltonian that are simulated in the
        stagger, the indices they act on, and whether they are hopping
        operators (all in the same order).

    Notes:
        The "staggers" used here are the left (parity=False) and right
        (parity=True) staggers detailed in Kivlichan et al., "Quantum
        Simulation of Electronic Structure with Linear Depth and
        Connectivity", arxiv:1711.04789. As such, the Hamiltonian must be
        in the form discussed in that paper. This constrains it to have
        only hopping terms (i^ j + j^ i) and potential terms which are
        products of at most two number operators (n_i or n_i n_j).
    """
    terms_in_layer = []
    indices_in_layer = []
    is_hopping_operator_in_layer = []

    n_qubits = count_qubits(hamiltonian)

    # A single round of odd-even transposition sort.
    for i in range(parity, n_qubits - 1, 2):
        # Always keep the max on the left to avoid having to normal order.
        left = max(input_ordering[i], input_ordering[i + 1])
        right = min(input_ordering[i], input_ordering[i + 1])

        # Calculate the hopping operators in the Hamiltonian.
        left_hopping_operator = FermionOperator(
            ((left, 1), (right, 0)),
            hamiltonian.terms.get(((left, 1), (right, 0)), 0.0))
        right_hopping_operator = FermionOperator(
            ((right, 1), (left, 0)),
            hamiltonian.terms.get(((right, 1), (left, 0)), 0.0))

        # Calculate the two-number operator l^ r^ l r in the Hamiltonian.
        two_number_operator = FermionOperator(
            ((left, 1), (right, 1), (left, 0), (right, 0)),
            hamiltonian.terms.get(
                ((left, 1), (right, 1), (left, 0), (right, 0)), 0.0))

        if not external_potential_at_end:
            # Calculate the left number operator, left^ left.
            left_number_operator = FermionOperator(
                ((left, 1), (left, 0)),
                hamiltonian.terms.get(((left, 1), (left, 0)), 0.0))

            # Calculate the right number operator, right^ right.
            right_number_operator = FermionOperator(
                ((right, 1), (right, 0)),
                hamiltonian.terms.get(((right, 1), (right, 0)), 0.0))

            # Divide single-number terms by n_qubits-1 to avoid over-accounting
            # for the interspersed rotations. Each qubit is swapped n_qubits-1
            # times total.
            left_number_operator /= (n_qubits - 1)
            right_number_operator /= (n_qubits - 1)

        else:
            left_number_operator = FermionOperator.zero()
            right_number_operator = FermionOperator.zero()

        # If the overall hopping operator isn't close to zero, append it.
        # Include the indices it acts on and that it's a hopping operator.
        if not (left_hopping_operator +
                right_hopping_operator) == FermionOperator.zero():
            terms_in_layer.append(left_hopping_operator +
                                  right_hopping_operator)
            indices_in_layer.append(set((left, right)))
            is_hopping_operator_in_layer.append(True)

        # If the overall number operator isn't close to zero, append it.
        # Include the indices it acts on and that it's a number operator.
        if not (two_number_operator + left_number_operator +
                right_number_operator) == FermionOperator.zero():
            terms_in_layer.append(two_number_operator + left_number_operator +
                                  right_number_operator)
            terms_in_layer[-1].compress()

            indices_in_layer.append(set((left, right)))
            is_hopping_operator_in_layer.append(False)

        # Modify the current Jordan-Wigner canonical ordering in-place.
        input_ordering[i], input_ordering[i + 1] = (input_ordering[i + 1],
                                                    input_ordering[i])

    return terms_in_layer, indices_in_layer, is_hopping_operator_in_layer


def ordered_low_depth_terms_no_info(hamiltonian):
    """Give terms from Hamiltonian in dictionary output order.

    Args:
        hamiltonian (FermionOperator): The Hamiltonian.

    Returns:
        A list of terms from the Hamiltonian in simulated order.

    Notes:
        Assumes the Hamiltonian is in the form discussed in Kivlichan
        et al., "Quantum Simulation of Electronic Structure with Linear
        Depth and Connectivity", arxiv:1711.04789. This constrains the
        Hamiltonian to have only hopping terms (i^ j + j^ i) and potential
        terms which are products of at most two number operators (n_i or
        n_i n_j).
    """
    count_qubits(hamiltonian)
    hamiltonian = normal_ordered(hamiltonian)
    terms = []

    for operators, coefficient in hamiltonian.terms.items():
        terms += [FermionOperator(operators, coefficient)]

    return terms



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests  jordan_wigner.py."""
import unittest

from openfermion.ops.operators import (BosonOperator, FermionOperator,
                                       QuadOperator)
from openfermion.utils import is_hermitian

from openfermion.transforms.repconversions.weyl_ordering import (
    mccoy, symmetric_ordering, weyl_polynomial_quantization)


class McCoyTest(unittest.TestCase):

    def setUp(self):
        self.mode = 0
        self.op_a = 'q'
        self.op_b = 'p'

    def test_identity(self):
        m = 0
        n = 0
        res = mccoy(self.mode, self.op_a, self.op_b, m, n)
        expected = {tuple(): 1.0}
        self.assertEqual(res, expected)

    def test_single_operator(self):
        m = 1
        n = 0
        res = mccoy(self.mode, self.op_a, self.op_b, m, n)
        expected = {((0, self.op_a),): 1.0}
        self.assertEqual(res, expected)

        m = 0
        n = 1
        res = mccoy(self.mode, self.op_a, self.op_b, m, n)
        expected = {((0, self.op_b),): 1.0}
        self.assertEqual(res, expected)

    def test_squared(self):
        m = 2
        n = 0
        res = mccoy(self.mode, self.op_a, self.op_b, m, n)
        expected = {((0, self.op_a), (0, self.op_a)): 1.0}
        self.assertEqual(res, expected)

        m = 0
        n = 2
        res = mccoy(self.mode, self.op_a, self.op_b, m, n)
        expected = {((0, self.op_b), (0, self.op_b)): 1.0}
        self.assertEqual(res, expected)

    def test_mixed(self):
        m = 1
        n = 1
        res = mccoy(self.mode, self.op_a, self.op_b, m, n)
        expected = {
            ((0, self.op_a), (0, self.op_b)): 0.5,
            ((0, self.op_b), (0, self.op_a)): 0.5
        }
        self.assertEqual(res, expected)


class WeylQuantizationTest(unittest.TestCase):

    def test_weyl_empty(self):
        res = weyl_polynomial_quantization('')
        self.assertTrue(res == QuadOperator.zero())

    def test_weyl_one_term(self):
        op = QuadOperator('q0')
        res = weyl_polynomial_quantization('q0')
        self.assertTrue(res == op)

    def test_weyl_one_term_multimode(self):
        op = QuadOperator('q0 q1 p2 p3')
        res = weyl_polynomial_quantization('q0 q1 p2 p3')
        self.assertTrue(res == op)

    def test_weyl_two_term_same(self):
        op = QuadOperator('q0 q0')
        res = weyl_polynomial_quantization('q0^2')
        self.assertTrue(res == op)

    def test_weyl_non_hermitian(self):
        res = weyl_polynomial_quantization('q0 p0')
        expected = QuadOperator('q0 p0', 0.5) \
            + QuadOperator('p0 q0', 0.5)
        self.assertTrue(res == expected)
        self.assertTrue(is_hermitian(res))

        res = weyl_polynomial_quantization('q0^2 p0')
        expected = QuadOperator('q0 q0 p0', 0.5) \
            + QuadOperator('p0 q0 q0', 0.5)
        self.assertTrue(res == expected)
        self.assertTrue(is_hermitian(res))


class SymmetricOrderingTest(unittest.TestCase):

    def test_invalid_op(self):
        op = FermionOperator()
        with self.assertRaises(TypeError):
            _ = symmetric_ordering(op)

    def test_symmetric_empty(self):
        for op in (BosonOperator, QuadOperator):
            res = symmetric_ordering(op())
            self.assertTrue(res == op().zero())

    def test_symmetric_identity(self):
        for op in (BosonOperator, QuadOperator):
            res = symmetric_ordering(op(''), ignore_identity=False)
            self.assertTrue(res == op().identity())

        for op in (BosonOperator, QuadOperator):
            res = symmetric_ordering(op(''), ignore_identity=True)
            self.assertTrue(res == op().zero())

    def test_symmetric_one_term(self):
        op = BosonOperator('0^')
        res = symmetric_ordering(op)
        self.assertTrue(res == op)

        op = QuadOperator('q0')
        res = symmetric_ordering(op)
        self.assertTrue(res == op)

    def test_symmetric_one_term_multimode(self):
        op = BosonOperator('0^ 1^ 2 3')
        res = symmetric_ordering(op)
        self.assertTrue(res == op)

        op = QuadOperator('q0 q1 p2 p3')
        res = symmetric_ordering(op)
        self.assertTrue(res == op)

    def test_symmetric_two_term_same(self):
        op = BosonOperator('0^ 0^')
        res = symmetric_ordering(op)
        self.assertTrue(res == op)

        op = QuadOperator('q0 q0')
        res = symmetric_ordering(op)
        self.assertTrue(res == op)

    def test_symmetric_non_hermitian(self):
        op = BosonOperator('0^ 0')
        res = symmetric_ordering(op)
        expected = BosonOperator('0^ 0', 0.5) \
            + BosonOperator('0 0^', 0.5)
        self.assertTrue(res == expected)
        self.assertTrue(is_hermitian(res))

        op = BosonOperator('0^ 0', 0.5)
        res = symmetric_ordering(op, ignore_coeff=False)
        expected = BosonOperator('0^ 0', 0.25) \
            + BosonOperator('0 0^', 0.25)
        self.assertTrue(res == expected)
        self.assertTrue(is_hermitian(res))

        op = QuadOperator('q0 p0')
        res = symmetric_ordering(op)
        expected = QuadOperator('q0 p0', 0.5) \
            + QuadOperator('p0 q0', 0.5)
        self.assertTrue(res == expected)
        self.assertTrue(is_hermitian(res))

    def test_symmetric_non_hermitian_order(self):
        op1 = QuadOperator('q0 p0 q0')
        op2 = QuadOperator('q0 q0 p0')
        op3 = QuadOperator('p0 q0 q0')

        w1 = symmetric_ordering(op1)
        w2 = symmetric_ordering(op2)
        w3 = symmetric_ordering(op3)

        self.assertTrue(is_hermitian(w1))
        self.assertTrue(is_hermitian(w2))
        self.assertTrue(is_hermitian(w3))

        expected = QuadOperator('q0 q0 p0', 0.5) \
            + QuadOperator('p0 q0 q0', 0.5)
        self.assertTrue(w1 == expected)
        self.assertTrue(w2 == expected)
        self.assertTrue(w3 == expected)

    def test_symmetric_coefficient(self):
        coeff = 0.5 + 0.6j
        op = QuadOperator('q0 p0', coeff)
        res = symmetric_ordering(op, ignore_coeff=False)
        expected = QuadOperator('q0 p0', 0.5) \
            + QuadOperator('p0 q0', 0.5)
        self.assertTrue(res == coeff * expected)
        self.assertFalse(is_hermitian(res))



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Dict, Tuple
from types import ModuleType
import warnings


def wrap_module(module: ModuleType,
                deprecated_attributes: Dict[str, Tuple[str, str]]):
    """Wrap a module with deprecated attributes.

    Args:
        module: The module to wrap.
        deprecated_attributes: A dictionary from attribute name to pair of
            strings, where the first string gives the version that the attribute
            will be removed in, and the second string describes what the user
            should do instead of accessing this deprecated attribute.

    Returns:
        Wrapped module with deprecated attributes.
    """

    class Wrapped(ModuleType):

        __dict__ = module.__dict__

        def __getattr__(self, name):
            if name in deprecated_attributes:
                version, fix = deprecated_attributes[name]
                warnings.warn(
                    f'{name} was used but is deprecated.\n'
                    f'It will be removed in '
                    f'openfermion {version}.\n'
                    f'{fix}\n',
                    DeprecationWarning,
                    stacklevel=2)
            return getattr(module, name)

    return Wrapped(module.__name__, module.__doc__)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest
import cirq

import openfermion


@pytest.mark.parametrize('rads', [
    2 * np.pi, np.pi, 0.5 * np.pi, 0.25 * np.pi, 0.1 * np.pi, 0.0, -0.5 * np.pi
])
def test_crxxyy_unitary(rads):
    np.testing.assert_allclose(
        cirq.unitary(openfermion.CRxxyy(rads)),
        cirq.unitary(cirq.ControlledGate(openfermion.Rxxyy(rads))),
        atol=1e-8)


@pytest.mark.parametrize('rads', [
    2 * np.pi, np.pi, 0.5 * np.pi, 0.25 * np.pi, 0.1 * np.pi, 0.0, -0.5 * np.pi
])
def test_cryxxy_unitary(rads):
    np.testing.assert_allclose(
        cirq.unitary(openfermion.CRyxxy(rads)),
        cirq.unitary(cirq.ControlledGate(openfermion.Ryxxy(rads))),
        atol=1e-8)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""The fast fermionic Fourier transform."""

from typing import (Iterable, List, Sequence)

import numpy as np
from sympy.ntheory import factorint

import cirq
import cirq.contrib.acquaintance.permutation
from openfermion.circuits.gates import FSWAP
from .bogoliubov_transform import bogoliubov_transform


class _F0Gate(cirq.MatrixGate):
    r"""Two-qubit gate that performs fermionic Fourier transform of size 2.

    Realizes unitary gate :math:`F_0` that transforms Fermionic creation
    operators :math:`a_0^\dagger` and :math:`a_1^\dagger` according to:

    .. math::
        F_0^\dagger a_0^\dagger F_0 =
            {1 \over \sqrt{2}} (a_0^\dagger + a_1^\dagger)

    .. math::
        F_0^\dagger a_1^\dagger F_0 =
            {1 \over \sqrt{2}} (a_0^\dagger - a_1^\dagger) \, .

    This gate assumes JWT representation of fermionic modes which are big-endian
    encoded on consecutive qubits:
    :math:`a_0^\dagger \lvert 00 \rangle = \lvert 10 \rangle` and
    :math:`a_1^\dagger \lvert 00 \rangle = \vert 01 \rangle`.

    Internally, this leads to expansion of :math:`F_0^\dagger`:

    .. math::
        \langle 00 \rvert F_0^\dagger \lvert 00 \rangle = 1

    .. math::
        \langle 01 \rvert F_0^\dagger \lvert 01 \rangle =
            -{1 \over \sqrt{2}}

    .. math::
        \langle 10 \rvert F_0^\dagger \lvert 10 \rangle =
        \langle 10 \rvert F_0^\dagger \lvert 01 \rangle =
        \langle 01 \rvert F_0^\dagger \lvert 10 \rangle = {1 \over \sqrt{2}}

    .. math::
        \langle 11 \rvert F_0^\dagger \lvert 11 \rangle = -1 \, .
    """

    def __init__(self):
        """Initializes :math:`F_0` gate."""
        cirq.MatrixGate.__init__(self,
                                 np.array([[1, 0, 0, 0],
                                           [0, -2**(-0.5), 2**(-0.5), 0],
                                           [0, 2**(-0.5), 2**(-0.5), 0],
                                           [0, 0, 0, -1]]),
                                 qid_shape=(2, 2))

    def _circuit_diagram_info_(self, args: cirq.CircuitDiagramInfoArgs
                              ) -> cirq.CircuitDiagramInfo:
        if args.use_unicode_characters:
            symbols = 'F₀', 'F₀'
        else:
            symbols = 'F0', 'F0'
        return cirq.CircuitDiagramInfo(wire_symbols=symbols)


class _TwiddleGate(cirq.SingleQubitGate):
    r"""Gate that introduces arbitrary FFT twiddle factors.

    Realizes unitary gate :math:`\omega^{k\dagger}_n` that phases creation
    operator :math:`a^\dagger_x` according to:

    .. math::
        \omega^{k\dagger}_n a^\dagger_x \omega^k_n =
            e^{-2 \pi i {k \over n}} a^\dagger_x \, .

    Under JWT representation this is realized by appropriately rotated pauli Z
    gate acting on qubit x.
    """

    def __init__(self, k, n):
        """Initializes Twiddle gate.

        Args:
            k: Numerator appearing in the exponent.
            n: Denominator appearing in the exponent.
        """
        self.k = k
        self.n = n

    def _circuit_diagram_info_(self, args: cirq.CircuitDiagramInfoArgs
                              ) -> cirq.CircuitDiagramInfo:
        if args.use_unicode_characters:
            symbols = 'ω^{}_{}'.format(self.k, self.n),
        else:
            symbols = 'w^{}_{}'.format(self.k, self.n),
        return cirq.CircuitDiagramInfo(wire_symbols=symbols)

    def _decompose_(self, qubits: Iterable[cirq.Qid]):
        q, = qubits
        exponent = -2 * self.k / self.n
        yield cirq.ZPowGate(exponent=exponent, global_shift=0).on(q)


F0 = _F0Gate()


def ffft(qubits: Sequence[cirq.Qid]) -> cirq.OP_TREE:
    r"""Performs fast fermionic Fourier transform.

    Generates a circuit that performs fast fermionic Fourier transform (FFFT)
    which transforms a set of fermionic creation operators
    :math:`\hat{a}_n^\dagger`, :math:`n \in 1, 2, \dots, N` according to:

    .. math::
        \mathit{FFFT}^\dagger \tilde{a}_k^\dagger \mathit{FFFT} =
            {1 \over \sqrt{N}}
            \sum_{n=0}^{N-1} e^{-i {2\pi \over N} n k} \hat{a}^\dagger_n \, ,

    where :math:`\tilde{a}_k^\dagger` are transformed operators and :math:`N` is
    size of the input `qubits` sequence.

    This function assumes JWT representation of fermionic modes which are
    big-endian encoded on consecutive qubits:
    :math:`a_0^\dagger \lvert 0.. \rangle = \lvert 1.. \rangle`,
    :math:`a_1^\dagger \lvert 0.. \rangle = \vert 01.. \rangle`,
    :math:`a_2^\dagger \lvert 0.. \rangle = \vert 001.. \rangle`, :math:`\dots`.

    The gate count of generated circuit is :math:`\theta(N^2)`, generated
    circuit depth is :math:`\theta(N)` and distinct gates count is
    :math:`\theta(N_1^2 + N_2^2 + \dots + N_n^2)`, where
    :math:`N = N_1 N_2 \dots N_n` is prime decomposition of :math:`N`. In a case
    where :math:`N` is some power of 2, it reduces to :math:`\theta(\log(N))`.

    An equivalent circuit can be generated using the
    `openfermion.bogoliubov_transform` function with appropriately prepared
    `transformation_matrix` argument:

    .. testcode::

        import openfermion
        import numpy as np

        def ffft(qubits):
            def fourier_transform_matrix(size):
                root_of_unity = np.exp(-2j * np.pi / size)
                return np.array([[root_of_unity ** (k * n) for n in range(size)]
                                 for k in range(size)]) / np.sqrt(size)

            return openfermion.bogoliubov_transform(
                qubits, fourier_transform_matrix(len(qubits)))

    The advantage of circuit generated by the FFFT algorithm over the one
    created with the `bogoliubov_transform` is that a smaller variety of gates
    is created, which is :math:`O(N^2)` in case of pure `bogoliubov_transform`.
    This implementation of `FFFT` fall-backs to the `bogoliubov_transform` for
    the inputs where qubits length is prime.

    This implementation of FFFT is based on a generalized, composite size
    Cooley-Tukey algorithm and adapted to nearest neighbourhood connectivity.

    Args:
        qubits: Sequence of qubits that the FFFT circuit will be generated for.
            This sequence represents a sequence of consecutive creation
            operators under big-endian encoded JWT representation. The indices
            assignment is significant since it it used to define the FFFT
            operation itself. The input sequence is assumed to have nearest
            neighbourhood connectivity.

    Returns:
        Circuit that performs FFFT on given qubits.

    Raises:
        ValueError: When length of the input array is not a power of 2.
    """
    n = len(qubits)

    if n == 0:
        raise ValueError('Number of qubits is 0.')

    if n == 1:
        return []

    factors = [f for f, count in factorint(n).items() for _ in range(count)]
    return _ffft(qubits, factors)


def _ffft_prime(qubits: Sequence[cirq.Qid]) -> cirq.OP_TREE:

    def fft_matrix(n):
        unit = np.exp(-2j * np.pi / n)
        return np.array([[unit**(j * k) for k in range(n)] for j in range(n)
                        ]) / np.sqrt(n)

    n = len(qubits)

    if n == 2:
        return F0(*qubits)
    else:
        return bogoliubov_transform(qubits, fft_matrix(n))


def _ffft(qubits: Sequence[cirq.Qid], factors: List[int]) -> cirq.OP_TREE:
    if len(factors) == 1:
        return _ffft_prime(qubits)

    n = len(qubits)

    # ny is a first prime factor, nx is a product of all the remaining ones.
    factors_y = factors[:1]
    factors_x = factors[1:]
    ny = np.prod(factors_y)
    nx = n // ny

    permutation = [(i % ny) * nx + (i // ny) for i in range(n)]
    permutation_gate = _permute(qubits, permutation)

    operations = []

    # This follows composite case Cooley-Tukey algorithm. The first step is to
    # recursively conduct ny FFFTs of size nx, where each of the FFFTs is
    # performed on nx consecutive qubits that are ny indices apart in the
    # original sequence.
    # This operation shuffles input qubits so that they are grouped together
    # accordingly. When ny = 2, this operation shuffles qubits to place odd
    # qubits in the first half of a sequence, and even qubits in the second
    # half.
    operations.append(permutation_gate)

    # Performs ny recursive FFFTs, each of size nx.
    for y in range(ny):
        operations.append(_ffft(qubits[nx * y:nx * (y + 1)], factors_x))

    # The second part is to perform ny FFFTs of size nx on qubits which are
    # consecutive in the original sequence. To place qubits in a correct order
    # the original permutation operation is inverted and applied to the previous
    # result.
    operations.append(cirq.inverse(permutation_gate))

    # Performs the nx FFFTs, each of size ny.
    for x in range(nx):
        for y in range(1, ny):
            operations.append(_TwiddleGate(x * y, n).on(qubits[ny * x + y]))
        operations.append(_ffft(qubits[ny * x:ny * (x + 1)], factors_y))

    # The result of performing FFFT on k-th group of nx qubits is a new group
    # of fermionic operators with indices k, k + ny, k + 2ny, ... of the final
    # fermionic operators sequence. The first permutation is applied again,
    # which puts every qubit in its final position.
    operations.append(permutation_gate)

    return operations


def _permute(qubits: Sequence[cirq.Qid],
             permutation: List[int]) -> cirq.OP_TREE:
    """
    Generates a circuit which reorders Fermionic modes.

    JWT representation of Fermionic modes is assumed. This is just a wrapper
    around cirq.contrib.acquaintance.permutation.LinearPermutationGate which
    internally uses bubble sort algorithm to generate permutation gate.

    Args:
        qubits: Sequence of qubits to reorder. Line connectivity is assumed.
        permutation: The permutation function represented as a list that
            reorders the initial qubits. Specifically, if k-th element of
            permutation is j, then k-th qubit should become the j-th qubit after
            applying the circuit to the initial state.

    Return:
        Gate that reorders the qubits accordingly.
    """
    return cirq.contrib.acquaintance.permutation.LinearPermutationGate(
        len(qubits), {i: permutation[i] for i in range(len(permutation))},
        FSWAP).on(*qubits)



#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Tests for QuadOperator."""

import unittest

from openfermion.ops.operators.quad_operator import QuadOperator


class QuadOperatorTest(unittest.TestCase):

    def test_is_normal_ordered_empty(self):
        op = QuadOperator() * 2
        self.assertTrue(op.is_normal_ordered())

    def test_is_normal_ordered_number(self):
        op = QuadOperator('q2 p2') * -1j
        self.assertTrue(op.is_normal_ordered())

    def test_is_normal_ordered_reversed(self):
        self.assertFalse(QuadOperator('p2 q2').is_normal_ordered())

    def test_is_normal_ordered_q(self):
        self.assertTrue(QuadOperator('q11').is_normal_ordered())

    def test_is_normal_ordered_p(self):
        self.assertTrue(QuadOperator('p0').is_normal_ordered())

    def test_is_normal_ordered_different_indices(self):
        self.assertTrue(QuadOperator('p0 q5 q3 q2 q1').is_normal_ordered())

    def test_is_normal_ordered_multi(self):
        op = QuadOperator('p4 p3 q2 p2') + QuadOperator('p1 p2')
        self.assertTrue(op.is_normal_ordered())

    def test_is_normal_ordered_multiorder(self):
        op = QuadOperator('p4 p3 p2 p1') + QuadOperator('p3 p2')
        self.assertTrue(op.is_normal_ordered())

    def test_is_gaussian_QuadOperator(self):
        op = QuadOperator()
        self.assertTrue(op.is_gaussian())

        op = QuadOperator('')
        self.assertTrue(op.is_gaussian())

        op1 = QuadOperator('q0')
        self.assertTrue(op1.is_gaussian())

        op2 = QuadOperator('q0 q0')
        self.assertTrue(op2.is_gaussian())

        op3 = QuadOperator('p0')
        self.assertTrue(op3.is_gaussian())

        op4 = QuadOperator('p0 p0')
        self.assertTrue(op4.is_gaussian())

        op5 = QuadOperator('q0 p0')
        self.assertTrue(op5.is_gaussian())

        op6 = QuadOperator('q0 q0 q0')
        self.assertFalse(op6.is_gaussian())

        op = op1 + op2 + op3 + op4 + op5
        self.assertTrue(op.is_gaussian())
        op += op6
        self.assertFalse(op.is_gaussian())
